# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for performing index server related operations on the commcell

IndexServers, IndexServer and _Roles are 3 classes defined in this file

IndexServers:   Class for representing all the index servers associated with the commcell

IndexServer:    Class for a instance of a single index server of the commcell

_Roles:         Class for storing all the cloud role details

"IndexServerOSType" is the enum class used to represent os type of IS

IndexServers
============

    __init__()                          --  initialize object of IndexServers class associated with
                                            the commcell

    __str()                             --  returns all the index servers of the commcell

    __repr__()                          --  returns the string to represent the instance

    __len__()                           --  returns the number of index servers associated

    _get_index_servers()                --  gets all the index server associated with the commcell

    _response_not_success()             --  raise exception when response is not 200

    _get_all_roles()                    --  creates an instance of _Roles class

    has()                               --  returns whether the index server is present or not

    get()                               --  returns a IndexServer object for given cloud name

    create()                            --  creates a index server within the commcell

    delete()                            --  deletes a index server associated with commcell

    update_roles_data()                 --  fetches the cloud roles data from commcell

    get_properties()                    --  returns a dict of data of index server for the given
                                            cloud name

    refresh()                           --  refresh the index servers associated with commcell

    prune_orphan_datasources()          --  Deletes all the orphan datasources

IndexServers Attributes
-----------------------

    **all_index_servers**               --  returns the dictionary consisting of all the index
                                            servers associated with the commcell and there details

    **roles_data**                      --  returns the list of cloud roles details


IndexServer
===========

    __init()__                          --  initializes the object with the specified commcell
                                            object, index server name and the cloud id

    __repr__()                          --  returns the index server's name, the instance is
    associated with

    _get_cloud_id()                     --  gets the cloud id

    _get_properties()                   --  gets all the properties of the index server

    refresh()                           --  refresh all the properties of client

    update_roles_data()                 --  fetches the cloud roles data from commcell

    modify()                            --  to modify the index server node details

    change_plan()                       --  changes the plan of a given index server

    update_role()                       --  to update the roles assigned to cloud

    hard_commit                         --  do hard commit on specified index server solr core

    get_all_cores                       --  gets all the cores in index server

    _create_solr_query()                --  Create solr search query based on inputs provided

    execute_solr_query()                --  Creates solr url based on input and executes it on solr on given core

    get_index_node()                    --  returns an Index server node object for given node name

    get_os_info()                       --  returns the OS type for the Index server

    get_plan_info()                     --  Returns the plan information of the index server

    __form_field_query()                --  returns the query with the key and value passed

IndexServer Attributes
----------------------

    **properties**                      --  returns the properties of this index server

    **roles_data**                      --  returns all the available cloud roles data

    **host_name**                       --  returns the host name for the index server

    **internal_cloud_name**             --  returns the internal cloud name

    **client_name**                     --  returns the client name for index server

    **server_url**                      --  returns the content indexing server url

    **type**                            --  returns the type of the index server

    **base_port**                       --  returns the base port of this index server

    **client_id**                       --  returns the client id for this index server

    **roles**                           --  returns the array of roles installed
                                            with the index server within the commcell

    **cloud_id**                        --  returns the cloud id of the index server

    **server_type**                     --  returns the server type of the index server

    **engine_name**                     --  returns the engine name that is index server name

    **index_server_client_id**          --  returns the index server client id

    **role_display_name**               --  display name of roles

    **is_cloud**                        --  returns boolean True if the Index server is cloud else returns False

    **node_count**                      --  returns the number of Index server nodes

    **os_info**                         --  returns the OS type for the Index server

    **plan_name**                       --  Returns the plan name associated with index server


IndexNode
=========

    __init__()                          --  initializes the class with commcell object
                                            Index server cloud id and Node client name

    refresh()                           --  refreshes the attributes

    modify()                            --  to modify the index server node details

IndexNode Attributes
--------------------

    **node_name**                       --  returns Index server node client name

    **node_id**                         --  returns Index server node client id

    **solr_port**                       --  returns port number Solr is running on the\
                                            Index server node

    **solr_url**                        --  returns Solr URL for Index server node

    **roles**                           --  returns the array of roles installed
                                            with the index server within the commcell

    **index_location**                  --  returns Index directory for the Index server Node

    **jvm_memory**                      --  returns Solr JVM memory for the Index server Node

_Roles
======

    __init__()                          --  initializes the class with commcell object

    refresh()                           --  refreshes the attributes

    _get_all_roles()                    --  fetches the cloud roles data from commcell

    get_role_id()                       --  returns role id for given role name

    update_roles_data()                 --  fetches the cloud roles data from commcell

_Roles Attributes
-----------------

    **roles_data**                      --  returns the list of details of all cloud roles
    """

from copy import deepcopy
import enum
from .exception import SDKException
from .datacube.constants import IndexServerConstants


class IndexServers(object):
    """Class for representing all the index servers associated with the commcell."""

    def __init__(self, commcell_object):
        """Initialize object of the IndexServers class.

            Args:
                commcell_object (object)  --  instance of the Commcell class

            Returns:
                object - instance of the IndexServers class
        """
        self._commcell_object = commcell_object
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._update_response_ = commcell_object._update_response_
        self._all_index_servers = None
        self._roles_obj = None
        self.refresh()

    def __str__(self):
        """Representation string consisting of all index servers of the commcell.

                Returns:
                    str - string of all the index servers with different roles associated
                    with the commcell
        """
        representation_string = '{:^5}\t{:^20}\n\n'.format('S. No.', 'IS Name')
        index = 1
        for index_server in self._all_index_servers:
            representation_string += '{:^5}\t{:^20}\n'.format(
                index, index_server['engineName'])
            index += 1
        return representation_string

    def __repr__(self):
        """Representation string for the instance of the IndexServers class."""
        return "IndexServers class instance for Commcell: '{0}'".format(
            self._commcell_object.commserv_name
        )

    def __len__(self):
        """Returns the number of the index servers associated with the commcell"""
        return len(self._all_index_servers)

    def _response_not_success(self, response):
        """Helper method to raise exception when response is not 200 (ok)

                Raises:
                    SDKException:
                        Response was not success
        """
        raise SDKException(
            'Response',
            '101',
            self._update_response_(
                response.text))

    def _get_index_servers(self):
        """Method to retrieve all the index server available on commcell.

            Raises:
                SDKException:
                    Failed to get the list of analytics engines

                    Response was not success
        """
        flag, response = self._cvpysdk_object.make_request(
            'GET', self._services['GET_ALL_INDEX_SERVERS'])
        if flag:
            if response.json() and 'listOfCIServer' in response.json():
                for item in response.json()['listOfCIServer']:
                    if item['cloudID'] in self._all_index_servers:
                        # Add only unique roles to list
                        if item['version'] not in self._all_index_servers[item['cloudID']]['version']:
                            self._all_index_servers[item['cloudID']]['version'].append(item['version'])
                        # check whether we have populated node details earlier. if not, add it to
                        # exisitng respective fields
                        if item['clientName'] not in self._all_index_servers[item['cloudID']]['clientName']:

                            self._all_index_servers[item['cloudID']]['clientId'].append(item['clientId'])
                            self._all_index_servers[item['cloudID']]['clientName'].append(item['clientName'])
                            self._all_index_servers[item['cloudID']]['hostName'].append(item['hostName'])
                            self._all_index_servers[item['cloudID']]['cIServerURL'].append(item['cIServerURL'])
                            self._all_index_servers[item['cloudID']]['basePort'].append(item['basePort'])

                    else:
                        item['version'] = [item['version']]
                        item['clientId'] = [item['clientId']]
                        item['clientName'] = [item['clientName']]
                        item['hostName'] = [item['hostName']]
                        item['cIServerURL'] = [item['cIServerURL']]
                        item['basePort'] = [item['basePort']]
                        self._all_index_servers[item['cloudID']] = item
            else:
                self._all_index_servers = {}
        else:
            self._response_not_success(response)

    def _get_all_roles(self):
        """Creates an instance of _Roles class and adds it to the IndexServer class"""
        self._roles_obj = _Roles(self._commcell_object)

    @property
    def all_index_servers(self):
        """Returns the details of all the index server for associated commcell.

                Returns:
                    dict - dictionary consisting details of all the index servers
                    associated with commcell
                    Sample - {
                                <cloud_id_1>   :
                                    {
                                        "engineName" : <property_value>,
                                        "internalCloudName" : <property_value>,
                                        ...
                                    },
                                <cloud_id_2>   :
                                    {
                                        "engineName" : <property_value>,
                                        "cloudID" : <property_value>,
                                        ...
                                    }
                            }
        """
        return self._all_index_servers

    @property
    def roles_data(self):
        """Returns the details of all the cloud roles data

                Returns:
                    list - list of dictionary containing details of the cloud roles
        """
        return self._roles_obj.roles_data

    def refresh(self):
        """Refresh the properties of IndexServers class"""
        self._all_index_servers = {}
        self._get_index_servers()
        if not self._roles_obj:
            self._get_all_roles()

    def update_roles_data(self):
        """Synchronises all the cloud roles details with the commcell"""
        self._roles_obj.update_roles_data()

    def get_properties(self, cloud_name):
        """Returns all details of a index server with the cloud name

                Args:
                    cloud_name     (str)       --  cloud name of index server

                Returns:
                    dict        -   dict consisting details of the index server
        """
        for index_server in self._all_index_servers:
            if self._all_index_servers[index_server]['engineName'] == cloud_name:
                return self._all_index_servers[index_server]
        raise SDKException('IndexServers', '102')

    def has(self, cloud_name):
        """Returns True if the index server with given name is present in commcell.

                Args:
                    cloud_name     (str)       --  the engine name of index server

                Returns:
                    boolean     -   True if index server with given name as is_name
                    is associated with the commcell else returns False

                Raises:
                    SDKExecption:
                        Data type of the input(s) is not valid
        """
        if isinstance(cloud_name, str):
            for index_server in self._all_index_servers:
                if self._all_index_servers[index_server]["engineName"].lower() == cloud_name.lower():
                    return True
            return False
        raise SDKException('IndexServers', '101')

    def get(self, cloud_data):
        """Returns IndexServer object if a index server is found.

                Args:
                    cloud_data        (int/str)       --    cloud name or
                                                            cloud ID of index server

                Returns:
                    object            (IndexServer)   --  Instance on index server with
                    the engine name or cloud id as item

                Raises:
                    SDKException:
                        Index Server not found.

                        Data type of the input(s) is not valid.
        """
        if isinstance(cloud_data, int):
            if cloud_data in self._all_index_servers:
                return IndexServer(
                    self._commcell_object,
                    self._all_index_servers[cloud_data]['engineName'],
                    cloud_data)
            SDKException('IndexServers', '102')
        elif isinstance(cloud_data, str):
            name = cloud_data.lower()
            for itter in self._all_index_servers:
                if self._all_index_servers[itter]['engineName'].lower(
                ) == name:
                    return IndexServer(
                        self._commcell_object,
                        self._all_index_servers[itter]['engineName'],
                        self._all_index_servers[itter]['cloudID'])
            raise SDKException('IndexServers', '102')
        raise SDKException('IndexServers', '101')

    def create(
            self,
            index_server_name,
            index_server_node_names,
            index_directory,
            index_server_roles,
            index_pool_name=None,
            is_cloud=False,
            cloud_param=None):
        """Creates an index server within the commcell

                Args:
                    index_server_node_names         (list)  --  client names for index server node
                    index_server_name               (str)   --  name for the index server
                    index_directory                 (list)  --  list of index locations for the index server
                                                                nodes respectively
                                                    For example:
                                                            [<path_1>] - same index location for all the nodes
                                                            [<path_1>, <path_2>, <path_3>] - different index
                                                    location for index server with 3 nodes
                    index_server_roles              (list)  --  list of role names to be assigned
                    index_pool_name                 (str)   --  name for the index pool to used by cloud index server
                    cloud_param                     (list)  --  list of custom parameters to be parsed
                                                    into the json for index server meta info
                                                    [
                                                        {
                                                            "name": <name>,
                                                            "value": <value>
                                                        }
                                                    ]
                    is_cloud            (bool)  --  if true then creates a cloud mode index server

                Raises:
                    SDKException:
                        Data type of the input(s) is not valid.

                        Response was not success.

                        Response was empty.
        """
        if not (isinstance(index_server_roles, list) and isinstance(index_server_node_names, list)
                and isinstance(index_server_name, str)):
            raise SDKException('IndexServers', '101')
        if isinstance(index_directory, str):
            index_directory = index_directory.split(",")
        node_count = len(index_server_node_names)
        index_directories_count = len(index_directory)
        if index_directories_count != 1 and index_directories_count != node_count:
            raise SDKException('IndexServers', '101')
        cloud_meta_infos = {
            'REPLICATION': '1',
            'LANGUAGE': '0'
        }
        node_meta_infos = {
            'PORTNO': IndexServerConstants.DEFAULT_SOLR_PORT,
            'JVMMAXMEMORY': IndexServerConstants.DEFAULT_JVM_MAX_MEMORY
        }
        role_meta_infos = {}
        req_json = deepcopy(IndexServerConstants.REQUEST_JSON)
        req_json['cloudInfoEntity'] = {
            'cloudName': index_server_name,
            'cloudDisplayName': index_server_name
        }
        if is_cloud:
            index_pool_obj = self._commcell_object.index_pools[index_pool_name]
            req_json['type'] = 5
            req_json['solrCloudInfo']['cloudPoolInfo'] = {
                'cloudId': int(index_pool_obj['pool_id'])
            }
            cloud_meta_infos['INDEXLOCATION'] = index_directory[0]
        for node_name_index in range(len(index_server_node_names)):
            node_name = index_server_node_names[node_name_index]
            location_index = node_name_index - (node_name_index//index_directories_count)
            node_obj = self._commcell_object.clients[node_name]
            node_data = {
                "opType": IndexServerConstants.OPERATION_ADD,
                "nodeClientEntity": {
                    "hostName": node_obj['hostname'],
                    "clientId": int(node_obj['id']),
                    "clientName": node_name
                },
                'nodeMetaInfos': [{
                    "name": "INDEXLOCATION",
                    "value": index_directory[location_index]
                }]
            }
            for node_info in node_meta_infos:
                node_data['nodeMetaInfos'].append({
                    'name': node_info,
                    'value': str(node_meta_infos[node_info])
                })
            req_json['cloudNodes'].append(node_data)
        for role in index_server_roles:
            role_id = self._roles_obj.get_role_id(role)
            if not role_id:
                raise SDKException('IndexServers', '103')
            role_data = {
                "roleId": role_id,
                "roleName": role,
                "operationType": IndexServerConstants.OPERATION_ADD,
                'roleMetaInfos': []
            }
            for role_info in role_meta_infos:
                role_data['roleMetaInfos'].append({
                    'name': role_info,
                    'value': role_meta_infos[role_info]
                })
            req_json['solrCloudInfo']['roles'].append(role_data)
        if cloud_param:
            for param in cloud_param:
                if param['name'] in cloud_meta_infos:
                    del cloud_meta_infos[param['name']]
                req_json['cloudMetaInfos'].append(param)
        for cloud_info in cloud_meta_infos:
            req_json['cloudMetaInfos'].append({
                'name': cloud_info,
                'value': cloud_meta_infos[cloud_info]
            })
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._services['CLOUD_CREATE'], req_json)
        if flag:
            if response.json():
                error_code = response.json()['genericResp']['errorCode']
                error_string = response.json()['genericResp']['errorMessage']
                if error_code == 0:
                    self.refresh()
                    self._commcell_object.clients.refresh()
                    self._commcell_object.datacube.refresh_engine()
                else:
                    o_str = 'Failed to create Index Server. Error: "{0}"'.format(
                        error_string)
                    raise SDKException('IndexServers', '102', o_str)
            else:
                raise SDKException('Response', '102')
        else:
            self._response_not_success(response)

    def delete(self, cloud_name):
        """Deletes / removes an index server from the commcell

                Args:
                    cloud_name      (str)   --  cloud name of index server
                    to be removed from the commcell

                Raises:
                    SDKException:
                        Data type of the input(s) is not valid.

                        Response was not success.

                        Response was empty.
        """
        if not isinstance(cloud_name, str):
            raise SDKException('IndexServers', '101')
        cloud_id = self.get(cloud_name).cloud_id
        req_json = deepcopy(IndexServerConstants.REQUEST_JSON)
        req_json["opType"] = IndexServerConstants.OPERATION_DELETE
        req_json['cloudInfoEntity']['cloudId'] = cloud_id
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._services['CLOUD_DELETE'], req_json
        )
        if flag:
            if response.json() and 'genericResp' in response.json() \
                    and 'errorCode' not in response.json()['genericResp']:
                self.refresh()
                self._commcell_object.clients.refresh()
                self._commcell_object.datacube.refresh_engine()
                return
            if response.json() and 'genericResp' in response.json():
                raise SDKException(
                    'Response', '102', response.json()['genericResp'].get(
                        'errorMessage', ''))
            raise SDKException('Response', '102')
        self._response_not_success(response)

    def prune_orphan_datasources(self):
        """Deletes all the orphan datasources
            Raises:
                SDKException:
                    if failed to prune the orphan datasources

                    If response is empty

                    if response is not success
        """
        prune_datasource = self._services['PRUNE_DATASOURCE']
        request_json = IndexServerConstants.PRUNE_REQUEST_JSON
        flag, response = self._cvpysdk_object.make_request(
            'POST', prune_datasource, request_json)
        if flag:
            if response.json():
                error_code = response.json().get('errorCode', 0)
                if error_code != 0:
                    raise SDKException('IndexServers', '104', 'Failed to prune orphan datasources')
                return
            raise SDKException('Response', '102')
        raise SDKException('Response', '101', self._update_response_(response.text))


class IndexServerOSType(enum.Enum):
    """Enum class for Index Server OS Type"""
    WINDOWS = "Windows"
    UNIX = "Unix"
    MIXED = "Mixed"


class IndexServer(object):
    """Class for performing index server operations for a specific index server"""

    def __init__(self, commcell_obj, name, cloud_id=None):
        """Initialize the IndexServer class instance.

            Args:
                commcell_obj    (object)        --  instance of the Commcell class

                name            (str)           --  name of the index server

                cloud_id        (int)           --  cloud id of the index server
                    default: None

            Returns:
                object - instance of the IndexServer class
        """
        self._engine_name = name
        self._commcell_obj = commcell_obj
        self._cvpysdk_object = self._commcell_obj._cvpysdk_object
        self._services = self._commcell_obj._services
        if cloud_id:
            self._cloud_id = cloud_id
        else:
            self._cloud_id = self._get_cloud_id()
        self._properties = None
        self._roles_obj = None
        self.plan_info = None
        self.os_type = None
        self.refresh()

    def __repr__(self):
        """String representation of the instance of this class."""
        return 'IndexServer class instance for index server: "{0}"'.format(
            self._engine_name)

    def _get_cloud_id(self):
        """Get the cloud id for the index server

                Returns:
                    int - cloud id for the index server
        """
        return self._commcell_obj.index_servers.get(self._engine_name).cloud_id

    def _get_properties(self):
        """Get the properties of the index server"""
        self._properties = self._commcell_obj.index_servers.get_properties(
            self._engine_name)

    def refresh(self):
        """Refresh the index server properties"""
        self._commcell_obj.index_servers.refresh()
        self._get_properties()
        if self.os_type is None:
            self.os_type = self.get_os_info()
        if not self._roles_obj:
            self._roles_obj = _Roles(self._commcell_obj)
        if self.plan_info is None:
            self.plan_info = self.get_plan_info()

    def update_roles_data(self):
        """Synchronize the cloud roles data with the commcell"""
        self._roles_obj.update_roles_data()

    def modify(self, index_location, node_name, node_params):
        """Modifies the properties of an index server

            Args:
                index_location      (str)       --  index server data directory
                node_name           (str)       --  index server node name
                node_params         (dict)      --  parameters to be passed
                                                    [
                                                        {
                                                            "name" : <property_name>,
                                                            "value" : <property_value>
                                                        }
                                                    ]
            Raises:
                SDKException:
                    Response was not success.
                    Response was empty.
        """
        json_req = deepcopy(IndexServerConstants.REQUEST_JSON)
        json_req['opType'] = IndexServerConstants.OPERATION_EDIT
        json_req['cloudNodes'] = [{
            "opType": IndexServerConstants.OPERATION_EDIT,
            "nodeClientEntity": {
                "clientId": int(self._commcell_obj.clients.get(node_name).client_id)
            },
            "nodeMetaInfos": [
                {
                    "name": "INDEXLOCATION",
                    "value": index_location
                }
            ]
        }]
        json_req['cloudInfoEntity']['cloudId'] = self.cloud_id
        for param in node_params:
            json_req['cloudNodes'][0]['nodeMetaInfos'].append(param)
        flag, response = self._cvpysdk_object.make_request(
            "POST", self._services['CLOUD_MODIFY'], json_req)
        if flag:
            if response.json():
                if 'cloudId' in response.json():
                    self.refresh()
                    return
            raise SDKException('Response', '102')
        raise SDKException('Response', '101')

    def change_plan(self, plan_name):
        """Modifies the plan used by an index server

            Args:
                plan_name      (str)       --  Name of the plan to be used for the index server
            Raises:
                SDKException:
                    Response was not success.
                    Response was empty.
                    if plan with given name doesn't exist
        """
        if not self._commcell_obj.plans.has_plan(plan_name):
            raise SDKException(
                'Plan', '102', f"Plan with name [{plan_name}] doesn't exist")
        request_json = {
            "opType": IndexServerConstants.OPERATION_EDIT,
            "type": 1,
            "planInfo": {
                "planId": int(self._commcell_obj.plans.get(plan_name).plan_id)
            },
            "cloudInfoEntity": {
                "cloudId": self.cloud_id
            }
        }
        flag, response = self._cvpysdk_object.make_request(
            "POST", self._services['CLOUD_MODIFY'], request_json)
        if flag:
            if response.json():
                if 'cloudId' in response.json():
                    self.refresh()
                    return
            raise SDKException('Response', '102')
        raise SDKException('Response', '101')

    def update_role(self, props=None):
        """Updates a role of an Index Server

            Args:
                props               (list)  --  array of dictionaries
                consisting details of the roles such as role name
                and operation type.
                                            [{
                                                "roleName": <name>          (str)
                                                "operationType": 1 or 2     (int)
                                                    1 for adding a role
                                                    2 for removing a role
                                            }]

            Raises:
                SDKException:
                    if response is empty

                    if response is not success

        """
        json_req = {"cloudId": self.cloud_id, "roles": []}
        if props:
            for prop in props:
                role_id = self._roles_obj.get_role_id(prop['roleName'])
                if not role_id:
                    raise SDKException('IndexServers', '103')
                prop['roleId'] = role_id
                json_req['roles'].append(prop)
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._services['CLOUD_ROLE_UPDATE'], json_req
        )
        if flag:
            if response.json() and 'errorCode' in response.json():
                if response.json()['errorCode'] == 0:
                    self.refresh()
                    return
            raise SDKException('Response', '102')
        raise SDKException('Response', '101')

    def hard_commit(self, core_name):
        """do hard commit for the given core name on index server

                    Args:

                        core_name               (str)  --  name of the solr core

                    Returns:
                        None

                    Raises:
                        SDKException:

                            if input data is not valid

                            if index server is cloud, not implemented error

                            if response is empty

                            if response is not success
        """
        if not isinstance(core_name, str):
            raise SDKException('IndexServers', '101')
        if self.is_cloud:
            raise SDKException('IndexServers', '104', "Not implemented for solr cloud")
        baseurl = f"{self.server_url[0]}/solr/{core_name}/update?commit=true"
        flag, response = self._cvpysdk_object.make_request("GET", baseurl)
        if flag and response.json():
            if 'error' in response.json():
                raise SDKException('IndexServers', '104', "Hard commit returned error")
            if 'responseHeader' in response.json():
                commitstatus = str(response.json()['responseHeader']['status'])
                if int(commitstatus) != 0:
                    raise SDKException('IndexServers', '104', "Hard commit returned bad status")
                return
        raise SDKException('IndexServers', '104', "Something went wrong with hard commit")

    def get_all_cores(self, client_name=None):
        """gets all cores & core details from index server

                Args:
                    client_name     (str)       --  name of the client node
                        ***Applicable only for solr cloud mode or multi node Index Server***

                Returns:
                    (list,dict)     -- list containing core names
                                    -- dict containing details about cores

                Raises:

                    SDKException:

                        if input data is not valid

                        if client name is not passed for index server cloud

                        if response is not success

                        if response is empty

        """
        server_url = self.server_url[0]
        if self.is_cloud or len(self.client_name) > 1:
            if client_name is None:
                raise SDKException('IndexServers', '104', 'Client name param missing for solr cloud')
            if client_name not in self.client_name:
                raise SDKException('IndexServers', '104', 'client name not found in this index server cloud')
            server_url = self.server_url[self.client_name.index(client_name)]
        core_names = []
        baseurl = f"{server_url}/solr/admin/cores"
        flag, response = self._cvpysdk_object.make_request("GET", baseurl)
        if flag and response.json():
            if 'error' in response.json():
                raise SDKException('IndexServers', '104', "Unable to get core names from index server")
            if 'status' in response.json():
                for core in response.json()['status']:
                    core_names.append(core)
                return core_names, response.json()['status']
        raise SDKException('IndexServers', '104', "Something went wrong while getting core names")

    def _create_solr_query(self, select_dict=None, attr_list=None, op_params=None):
        """Method to create the solr query based on the params
            Args:
                select_dict     (dictionary)     --  Dictionary containing search criteria and value
                                                     Acts as 'q' field in solr query

                attr_list       (set)            --  Column names to be returned in results.
                                                     Acts as 'fl' in solr query

                op_params       (dictionary)     --  Other params and values for solr query
                                                        (Ex: start, rows)

            Returns:
                The solr url based on params

            Raises:
                SDKException:

                        if failed to form solr query
        """
        try:
            search_query = f'q='
            simple_search = 0
            if select_dict:
                for key, value in select_dict.items():
                    if isinstance(key, tuple):
                        if isinstance(value, list):
                            search_query += f'({key[0]}:{str(value[0])}'
                            for val in value[1:]:
                                search_query += f' OR {key[0]}:{str(val)}'
                        else:
                            search_query += f'({key[0]}:{value}'
                        for key_val in key[1:]:
                            if isinstance(value, list):
                                search_query += f' OR {key_val}:{str(value[0])}'
                                for val in value[1:]:
                                    search_query += f' OR {key_val}:{str(val)}'
                            else:
                                search_query += f' OR {key_val}:{value}'
                        search_query += ') AND '
                    elif isinstance(value, list):
                        search_query += f'({key}:{str(value[0])}'
                        for val in value[1:]:
                            search_query += f' OR {key}:{str(val)}'
                        search_query += ") AND "
                    elif key == "keyword":
                        search_query += "(" + value + ")"
                        simple_search = 1
                        break
                    else:
                        search_query = search_query + f'{key}:{str(value)} AND '
                if simple_search == 0:
                    search_query = search_query[:-5]
            else:
                search_query += "*:*"

            field_query = ""
            if attr_list:
                field_query = "&fl="
                for item in attr_list:
                    field_query += f'{str(item)},'
                field_query = field_query[:-1]
            if attr_list and 'content' in attr_list:
                field_query = f"{field_query}&exclude=false"

            ex_query = ""
            if not op_params:
                op_params = {'wt': "json"}
            else:
                op_params['wt'] = "json"
            for key, values in op_params.items():
                if isinstance(values, list):
                    for value in values:
                        ex_query += self.__form_field_query(key, value)
                else:
                    ex_query += self.__form_field_query(key, values)
            final_url = f'{search_query}{field_query}{ex_query}'
            return final_url
        except Exception as excp:
            raise SDKException('IndexServers', '104', f"Something went wrong while creating solr query - {excp}")

    def execute_solr_query(
            self,
            core_name,
            solr_client=None,
            select_dict=None,
            attr_list=None,
            op_params=None):
        """Creates solr url based on input and executes it on solr on given core/collection
            Args:

                core_name               (str)           --  Core name/collection name where we want to query

                solr_client             (str)           --  Index Server client name to execute solr query
                                                                Default : None (picks first client on index server)

                select_dict             (dictionary)    --  Dictionary containing search criteria and
                                                            value. Acts as 'q' field in solr query

                        Example :

                            1. General Criteria to filter results              -   {"jid": 1024, "datatype": 2,clid: 2}
                            2. Keyword Searches on solr                        -   {'keyword': 'SearchKeyword'}
                            3. For multiple value searches on single field     -   {'cvowner': ['xxx','yyy']}
                            4. For single value searches on multiple fields    -   {('cvowner','cvreaddisp') : 'xxx'}

                attr_list               (set)           --  Column names to be returned in results.
                                                                Acts as 'fl' in solr query

                        Example (For Exchange Mailbox IDA, below fields are there in solr) :
                                    {
                                     'msgclass',
                                     'ccsmtp',
                                     'fmsmtp',
                                     'folder'
                                   }

                op_params               (dictionary)    --  Other params and values for solr query. Do not
                                                            mention 'wt' param as it is always json

                                                            Example : {"rows": 0}

            Returns:
                content of the response

            Raises:
                SDKException:

                        if unable to send request

                        if response is not success
        """
        solr_url = f"solr/{core_name}/select?{self._create_solr_query(select_dict, attr_list, op_params)}"
        if solr_client is None:
            solr_url = f"{self.server_url[0]}/{solr_url}"
        else:
            if solr_client not in self.client_name:
                raise SDKException('IndexServers', '104', 'client name not found in this index server')
            server_url = self.server_url[self.client_name.index(solr_client)]
            solr_url = f"{server_url}/{solr_url}"
        flag, response = self._cvpysdk_object.make_request("GET", solr_url)
        if flag and response.json():
            return response.json()
        raise SDKException('IndexServers', '104', "Something went wrong while querying solr")

    def get_index_node(self, node_name):
        """Returns an Index server node object for given node name
            Args:
                node_name           (str)   --  Index server node name

            Returns:
                IndexNode class object

            Raises:
                SDKException:

                        if node not found for the given node name

        """
        node_name = node_name.lower()
        if node_name in self.client_name:
            return IndexNode(self._commcell_obj, self.engine_name, node_name)
        raise SDKException("IndexServers", '104', 'Index server node not found')

    def get_plan_info(self):
        """Returns the plan information of the index server"""
        client = self._commcell_obj.clients.get(self.index_server_client_id)
        instance_props = client.properties.get("pseudoClientInfo", {}).get("distributedClusterInstanceProperties", {})
        plan_details = instance_props.get("clusterConfig",{}).get("cloudInfo", {}).get("planInfo", {})
        return plan_details

    def get_os_info(self):
        """Returns the OS type for the Index server"""

        nodes_name = self.client_name
        nodes = [self._commcell_obj.clients.get(node) for node in nodes_name]
        nodes_os_info = [node.os_info for node in nodes]
        if IndexServerOSType.WINDOWS.value.lower() in nodes_os_info[0].lower():
            for node in nodes_os_info[1:]:
                if IndexServerOSType.UNIX.value.lower() in node.lower():
                    return IndexServerOSType.MIXED.value
            return IndexServerOSType.WINDOWS.value
        else:
            for node in nodes_os_info[1:]:
                if IndexServerOSType.WINDOWS.value.lower() in node.lower():
                    return IndexServerOSType.MIXED.value
            return IndexServerOSType.UNIX.value

    def __form_field_query(self, key, value):
        """
        Returns the query with the key and value passed
        Args:
                key(str)    -- key for forming the query
                value(str)  -- value for forming the query
            Returns:
                query to be executed against solr
        """
        query = None
        if value is None:
            query = f'&{key}'
        else:
            query = f'&{key}={str(value)}'
        return query

    @property
    def plan_name(self):
        """Returns the plan name associated with index server"""
        return self.plan_info.get("planName")

    @property
    def os_info(self):
        """Returns the OS type for the Index server"""
        return self.os_type

    @property
    def is_cloud(self):
        """Returns true if the Index server is cloud and false if not"""
        return self.server_type == 5

    @property
    def nodes_count(self):
        """Returns the count of Index server nodes"""
        return len(self.client_id)

    @property
    def roles_data(self):
        """Returns the cloud roles data"""
        return self._roles_obj.roles_data

    @property
    def properties(self):
        """Returns the index server properties"""
        return self._properties

    @property
    def host_name(self):
        """Returns a list of host names of all index server nodes"""
        return self._properties[IndexServerConstants.HOST_NAME]

    @property
    def cloud_name(self):
        """Returns the internal cloud name of index server"""
        return self._properties[IndexServerConstants.CLOUD_NAME]

    @property
    def client_name(self):
        """Returns a list of client names of all index server nodes"""
        return self._properties[IndexServerConstants.CLIENT_NAME]

    @property
    def server_url(self):
        """Returns a list of Solr url of all index server nodes"""
        return self._properties[IndexServerConstants.CI_SERVER_URL]

    @property
    def type(self):
        """Returns the type of index server"""
        return self._properties[IndexServerConstants.TYPE]

    @property
    def base_port(self):
        """Returns a list of base ports of all index server nodes"""
        return self._properties[IndexServerConstants.BASE_PORT]

    @property
    def client_id(self):
        """Returns a list client ids of all index server nodes"""
        return self._properties[IndexServerConstants.CLIENT_ID]

    @property
    def roles(self):
        """Returns a list of roles of index server"""
        return self._properties[IndexServerConstants.ROLES]

    @property
    def role_display_name(self):
        """Returns the roles display name of index server"""
        role_disp_name = []
        for role_version in self.roles:
            for role in self.roles_data:
                if role_version == role['roleVersion']:
                    role_disp_name.append(role['roleName'])
                    break
        return role_disp_name

    @property
    def cloud_id(self):
        """Returns the cloud id of index server"""
        return self._properties[IndexServerConstants.CLOUD_ID]

    @property
    def server_type(self):
        """Returns the server type of index server"""
        return self._properties[IndexServerConstants.SERVER_TYPE]

    @property
    def engine_name(self):
        """Returns the engine name of index server"""
        return self._properties[IndexServerConstants.ENGINE_NAME]

    @property
    def index_server_client_id(self):
        """Returns the index server client id of index server"""
        return self._properties[IndexServerConstants.INDEX_SERVER_CLIENT_ID]


class IndexNode(object):
    """Class for Index server node object"""

    def __init__(self, commcell_obj, index_server_name, node_name):
        """Initialize the IndexNode class

            Args:
                commcell_obj        (object)    --  commcell object
                index_server_name   (int)       --  Index server name
                node_name           (str)       --  Index server node client name

        """
        self.commcell = commcell_obj
        self.index_server_name = index_server_name
        self.data_index = 0
        self.index_server = None
        self.index_node_name = node_name.lower()
        self.index_node_client = None
        self.index_client_properties = None
        self.refresh()

    def refresh(self):
        """Refresh the index node properties"""
        self.commcell.index_servers.refresh()
        self.index_server = self.commcell.index_servers.get(self.index_server_name)
        self.data_index = self.index_server.client_name.index(self.index_node_name)
        self.commcell.clients.refresh()
        self.index_node_client = self.commcell.clients.get(self.index_node_name)
        self.index_client_properties = self.index_node_client.properties['pseudoClientInfo']['indexServerProperties']

    @property
    def node_name(self):
        """Returns Index server node client name"""
        return self.index_server.client_name[self.data_index]

    @property
    def node_id(self):
        """Returns Index server node client id"""
        return self.index_server.client_id[self.data_index]

    @property
    def solr_port(self):
        """Returns port number Solr is running on the Index server node"""
        return self.index_server.base_port[self.data_index]

    @property
    def solr_url(self):
        """Returns Solr URL for Index server node"""
        return self.index_server.server_url[self.data_index]

    @property
    def roles(self):
        """Returns the array of roles installed with the index server within the commcell"""
        return self.index_server.role_display_name

    @property
    def index_location(self):
        """Returns Index directory for the Index server Node"""
        node_meta_infos = self.index_client_properties['nodeMetaInfos']
        for info in node_meta_infos:
            if info['name'] == 'INDEXLOCATION':
                return info['value']
        return None

    @property
    def jvm_memory(self):
        """Returns Solr JVM memory for the Index server Node"""
        node_meta_infos = self.index_client_properties['nodeMetaInfos']
        for info in node_meta_infos:
            if info['name'] == 'JVMMAXMEMORY':
                return info['value']
        return None

    @solr_port.setter
    def solr_port(self, port_no):
        """Setter to set the Solr port number for the node

            Args:
                port_no     (str)   --  Solr port number to be set for node

        """
        solr_port_param = deepcopy(IndexServerConstants.SOLR_PORT_META_INFO)
        solr_port_param['value'] = str(port_no)
        cloud_param = [solr_port_param]
        self.index_server.modify(self.index_location, self.index_node_name, cloud_param)
        self.refresh()

    @jvm_memory.setter
    def jvm_memory(self, memory):
        """Setter to set the Solr JVM memory for the node

                    Args:
                        memory      (str)   --  Solr JVM memory to be set for the node

        """
        solr_jvm_param = deepcopy(IndexServerConstants.SOLR_JVM_META_INFO)
        solr_jvm_param['value'] = str(memory)
        solr_port_param = deepcopy(IndexServerConstants.SOLR_PORT_META_INFO)
        solr_port_param['value'] = str(self.solr_port)
        cloud_param = [solr_jvm_param, solr_port_param]
        self.index_server.modify(self.index_location, self.index_node_name, cloud_param)
        self.refresh()


class _Roles(object):
    """Class for cloud roles data operations"""

    def __init__(self, commcell_object):
        """Initializes _Roles class with commcell object

            Args:
                commcell_object (object)    --  instance of Commcell class

            Returns:
                object  -   instance of _Roles class
        """
        self.commcell_object = commcell_object
        self._roles_data = None
        self.refresh()

    def refresh(self):
        """Refreshes the class data"""
        self._get_all_roles()

    def _get_all_roles(self):
        """Method to get all cloud roles details available on the commcell.

            Raises:
                SDKException:
                    Response was empty.

                    Response was not success.
        """
        flag, response = self.commcell_object._cvpysdk_object.make_request(
            "GET", self.commcell_object._services['GET_ALL_ROLES']
        )
        if flag:
            if response.json():
                if 'rolesInfo' in response.json():
                    self._roles_data = response.json()['rolesInfo']
                    return
            raise SDKException('Response', '102')
        raise SDKException('Response', '101')

    def get_role_id(self, role_name):
        """Method to get a cloud role id with given name

            Args:
                role_name       (str)   --  cloud role name of which role id has to be returned

            Returns:
                role_id         (int)   --  if role name is found in roles data then returns the id
                                            else returns None

        """
        for role_data in self._roles_data:
            if role_data['roleName'] == role_name:
                return role_data['roleId']
        return None

    def update_roles_data(self):
        """Synchronize the cloud role data with the commcell database"""
        self._get_all_roles()

    @property
    def roles_data(self):
        """Returns the list of dictionary of details of each cloud role"""
        return self._roles_data



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""
Main File for performing Sybase Subclient Operations

SybaseSubclient is the only class defined in this file.

SybaseSubclient :       Derived class from DatabaseSubclient Base class,
                        representing an Sybase subclient,
                        and to perform operations on that subclient

SybaseSubclient:

    __init__()                          --  initialise object of sybase
                                            subclient object associated
                                            with the specified instance

    _get_subclient_properties           --  get the all subclient related
                                            properties of this subclient

    _sybase_backup_request_json         --  Returns the JSON request to pass to
                                            the API as per the options selected by the user

    is_snapenabled()                    --  Check if intellisnap has been enabled
                                            in the subclient and sets it accordingly

    snap_engine()                       --  updates snap_engine for sybase subclient

    snap_proxy()                        --  updates proxy name from sybase snap operation

    use_dump_based_backup_copy()        --  updates the use of dump based method for backup copy

    dump_based_backup_copy_option()     --  updates subtype of dump based operation

    configured_instance()               --  updates configured instance name
                                            for dump based backup copy type 1

    auxiliary_sybase_server()           --  updates custom instance properties
                                            for dump based backup copy type 2

    content()                           --  update the content of
                                            the sybase  subclient

    backup()                            --  Run a backup job for the subclient


"""
from __future__ import unicode_literals
from ..subclient import Subclient
from ..exception import SDKException


class SybaseSubclient(Subclient):
    """
    Base class consisting of all the common properties and operations for a Sybase Subclient
    """

    def __init__(self, backupset_object, subclient_name, subclient_id=None):
        """
        Initialize Sybase Subclient Object

        Args:
            backupset_object  (object)  --  instance of the Backupset class

            subclient_name    (str)     --  name of the subclient

            subclient_id      (str)     --  id of the subclient
                                            default : None


        Returns :
            (object) - instance of the Sybase Subclient class

        """
        self._sybase_properties = {}
        self._snap_copy_info = None
        super(SybaseSubclient, self).__init__(
            backupset_object, subclient_name, subclient_id)

    def _get_subclient_properties(self):
        """
        Gets the subclient related properties of Sybase subclient
        """

        super(SybaseSubclient, self)._get_subclient_properties()
        if 'content' in self._subclient_properties:
            self._content = self._subclient_properties['content']

        self._snap_copy_info = self._commonProperties.get('snapCopyInfo')

    def _get_subclient_properties_json(self):
        """
        Get the all subclient related properties of this subclient.

           Returns:
                (dict) - all subclient properties put inside a dict

        """
        return {
            "subClientProperties":
                {
                    "proxyClient": self._proxyClient,
                    "subClientEntity": self._subClientEntity,
                    "content": self._content,
                    "commonProperties": self._commonProperties,
                    "contentOperationType": 1,
                    "snapCopyInfo": self._snap_copy_info
                }
        }

    def _sybase_backup_request_json(self,
                                    backup_level,
                                    do_not_truncate_log=False,
                                    sybase_skip_full_after_logbkp=False,
                                    create_backup_copy_immediately=False,
                                    backup_copy_type=2,
                                    directive_file=None):
        """
        Returns the JSON request to pass to the API as per the options selected by the user.

            Args:
               backup_level                     (list)  --  level of backup the user wish to run
                                                            Full / Incremental / Differential

               do_not_truncate_log              (bool)  --  Sybase truncate log option
                                                            for incremental backup
                                                            default : False

               sybase_skip_full_after_logbkp    (bool)  --  Sybase backup option for incremental
                                                            default : False

               create_backup_copy_immediately   (bool)  --  Sybase snap job needs
                                                            this backup copy operation
                                                            default : False

               backup_copy_type                 (int)   --  backup copy job to be launched
                                                            based on below two options
                                                            default : 2, possible values :
                                                            1 (USING_STORAGE_POLICY_RULE),
                                                            2( USING_LATEST_CYCLE)

               directive_file                   (str)   --  inputfile for ondemand backup
                                                            containing database list
                                                            default : None

            Returns:

                (dict) - JSON request to pass to the API

        """
        request_json = self._backup_json(backup_level, False, "BEFORE_SYNTH")
        sybase_options = {
            "doNotTruncateLog": do_not_truncate_log,
            "sybaseSkipFullafterLogBkp": sybase_skip_full_after_logbkp
        }

        if create_backup_copy_immediately:
            sub_opt = {"dataOpt":
                       {
                           "createBackupCopyImmediately": create_backup_copy_immediately,
                           "backupCopyType": backup_copy_type
                       }
                      }
            sybase_options.update(sub_opt)
        if self._commonProperties.get("onDemandSubClient", False):
            on_demand_input = {"onDemandInputFile":directive_file}
            sybase_options.update(on_demand_input)

        request_json["taskInfo"]["subTasks"][0]["options"]["backupOpts"].update(
            sybase_options
        )
        return request_json


    @property
    def is_snapenabled(self):
        """
        Getter to check whether the subclient has snap enabled

        Returns:
            (bool)  -    boolean value based on snap
                         status at subclient level

                True    -  returns Truee if snap is enabled on the subclient
                False   -  returns False if snap is not
                            enabled at subclient level

        """
        return self._snap_copy_info.get("isSnapBackupEnabled", False)

    @is_snapenabled.setter
    def is_snapenabled(self, value):
        """
        To set is snap enabled to true or false

        Args:
            value           (bool) --   to enable snap at subclient level or not

        """
        self._set_subclient_properties(
            "_snap_copy_info['isSnapBackupEnabled']", value)

    @property
    def snap_engine(self):
        """
        Getter to fetch snap_engine

        Returns:
            (str)     -  name of snap engine at subclient level

        """
        return self._snap_copy_info.get('snapToTapeSelectedEngine', {}).get('snapShotEngineName')

    @snap_engine.setter
    def snap_engine(self, engine_name):
        """
        To set snap engine name

        Args:
            engine_name           (str) --      name of snap engine
                                                for intellisnap

        """
        self._set_subclient_properties(
            "_snap_copy_info['snapToTapeSelectedEngine']['snapShotEngineName']", engine_name)

    @property
    def snap_proxy(self):
        """
        Getter to snap_proxy if set any

        Returns:
            (str)     --    name of proxy client used
                            for intellisnap operation

        """
        return self._snap_copy_info.get('snapToTapeProxyToUse', {}).get('clientName')

    @snap_proxy.setter
    def snap_proxy(self, proxy_name):
        """
        Setter for snap proxy name

        Args:
            proxy_name           (str) --   snap proxy name

        """
        self._set_subclient_properties(
            "_snap_copy_info['snapToTapeProxyToUse']['clientName']", proxy_name)

    @property
    def use_dump_based_backup_copy(self):
        """
        Getter to status of dumpbased backup copy

        Returns:
            (bool)     -    checks if dump based backup
                            copy is enabled or not

        """
        return self._snap_copy_info.get('useDumpBasedBackupCopy')

    @use_dump_based_backup_copy.setter
    def use_dump_based_backup_copy(self, dump_based):
        """
        To enable dump based backup copy

        Args:
            dump_based      (bool)     --      set true  to enable dump
                                               based backup copy option

        """
        self._set_subclient_properties(
            "_snap_copy_info['useDumpBasedBackupCopy']", dump_based)

    @property
    def dump_based_backup_copy_option(self):
        """
        Getter to fetch dumpbased backup copy option :
        1(configured instance), 2(custom new instance)

        Returns:
            (int)     -  returns 1 or 2 based type of sybase instance configured

        """
        return self._snap_copy_info.get("dumpBasedBackupCopyOption")

    @dump_based_backup_copy_option.setter
    def dump_based_backup_copy_option(self, dump_based_backup_copy_option):
        """
        Enable dump based backup copy

        Args:
            dump_based_backup_copy_option      (int) --  set 1(configured instance),
                                                             2(custom new instance)
        """
        self._set_subclient_properties(
            "_snap_copy_info['dumpBasedBackupCopyOption']", dump_based_backup_copy_option)

    @property
    def configured_instance(self):
        """
        Getter to fetch configured instance
        name if dump based backup copy option is 2
        Returns:
            (str)       -       string of configured instance
                                if dump based backup option is 1

        Raises:
            SDK Exception
                if dump based backup copy not enabled

                if dump based copy option is not 1

        """
        if self._snap_copy_info.get('useDumpBasedBackupCopy'):
            if self._snap_copy_info.get('dumpBasedBackupCopyOption') == 1:
                return self._snap_copy_info['configuredSybaseInstance']['instanceName']
            else:
                raise SDKException(
                    'Subclient', '102', "Invalid dump based copy option")
        else:
            raise SDKException('Subclient', '102',
                               "Dump based parameter is not available")

    @configured_instance.setter
    def configured_instance(self, instance_name):
        """
        Setter for configured instance
        name for dump based backup copy option

        Args:
            instance_name       (str) --    string of instance name
                                            to be used for dump based backup copy

        """

        self._set_subclient_properties(
            "_snap_copy_info['configuredSybaseInstance']['instanceName']", instance_name)

    @property
    def auxiliary_sybase_server(self):
        """
        Getter to fetch custom instance properties if dump based copy option is 2

        Returns:
            (dict)       -  dict of four properties
                            for custom instance
        Raises:
            SDK Exception
                if dump based backup copy not enabled

                if dump based copy option is not 2

        """

        if self.use_dump_based_backup_copy:
            if self.dump_based_backup_copy_option == 2:
                auxiliary_sybase_server = {
                    'sybaseHome': self._snap_copy_info.get('sybaseHome'),
                    'sybaseASE': self._snap_copy_info.get('sybaseASE'),
                    'sybaseOCS': self._snap_copy_info.get('sybaseOCS'),
                    'sybaseUser': self._snap_copy_info.get('sybaseUser', {}).get('userName')
                }
                return auxiliary_sybase_server
            else:
                raise SDKException(
                    'Subclient', '102', "Invalid dump based copy option set")
        else:
            raise SDKException('Subclient', '102',
                               "dump based parameter is not available")

    @auxiliary_sybase_server.setter
    def auxiliary_sybase_server(self, instance_properties):
        """
        Setter  custom instance properties if dump based copy option is 2

        Args:
            instance_properties     (dict)       --  dict of four properties
                                                     for custom instance

            Sample dict:
            instance_properties = {
                        'sybaseHome':sybase_home,
                        'sybaseASE':sybase_ase,
                        'sybaseOCS':sybase_ocs,
                        'sybaseUser':sybase_user
                }
        Raises:
            SDK Exception

                if None value in instance_properties

        """
        if None in instance_properties.values():
            raise SDKException(
                'Instance', '102', "One of the sybase custom instance parameter is None. Exiting")
        self._set_subclient_properties(
            "_snap_copy_info['sybaseHome']", instance_properties['sybaseHome'])
        self._set_subclient_properties(
            "_snap_copy_info['sybaseASE']", instance_properties['sybaseASE'])
        self._set_subclient_properties(
            "_snap_copy_info['sybaseOCS']", instance_properties['sybaseOCS'])
        self._set_subclient_properties("_snap_copy_info['sybaseUser']['userName']",
                                       instance_properties['sybaseUser'])

    @property
    def content(self):
        """Treats the subclient content as a property of the Subclient class."""
        subclient_content = self._content
        sybase_dblist = []
        for item in subclient_content:
            sybase_server_dict = item
            dbname = sybase_server_dict['sybaseContent']['databaseName']
            sybase_dblist.append(dbname)

        return sybase_dblist

    @content.setter
    def content(self, subclient_content):
        """
        Creates the list of content JSON to pass to the API to add a new Sybase Subclient
        with the content passed in subclient content.

            Args:
                subclient_content (list)  --  list of the content to add to the subclient

            Returns:
                (list) -    list of the appropriate JSON
                            for an agent to send to the POST Subclient API

        """
        content_new = []
        for dbname in subclient_content:
            sybase_server_dict = {"sybaseContent": {"databaseName": dbname}}
            content_new.append(sybase_server_dict)
        self._set_subclient_properties("_content", content_new)

    def backup(self,
               backup_level=r'full',
               do_not_truncate_log=False,
               sybase_skip_full_after_logbkp=False,
               create_backup_copy_immediately=False,
               backup_copy_type=2,
               directive_file=None):
        """
        Performs backup on sybase subclient

        Args:
            backup_level                            (str)   --  Level of backup.
                                                                full|incremental|differential
                                                                default: full

            do_not_truncate_log                     (bool)  --  Sybase truncate log option
                                                                for incremental backup
                                                                default : False

            sybase_skip_full_after_logbkp           (bool)  --  Sybase backup option for incremental
                                                                default : False

            create_backup_copy_immediately          (bool)  --  Sybase snap job needs
                                                                this backup copy operation
                                                                default : False

            backup_copy_type                        (int)   --  backup copy job to be launched
                                                                based on below two options
                                                                default : 2, possible values :
                                                                1 (USING_STORAGE_POLICY_RULE),
                                                                2( USING_LATEST_CYCLE)

            directive_file                          (str)   --  input file for ondemand backup
                                                                containing database list
                                                                default : None

        Returns:
            (object) - instance of Job class

        Raises:
            SDKException:
                if backup level is incorrect

                if response is empty

                if response does not succeed

        """
        if backup_level.lower() not in ['full', 'incremental', 'differential']:
            raise SDKException(r'Subclient', r'103')

        if backup_level.lower() == 'incremental':
            do_not_truncate_log = do_not_truncate_log
            sybase_skip_full_after_logbkp = sybase_skip_full_after_logbkp
        else:
            do_not_truncate_log = False
            sybase_skip_full_after_logbkp = False

        if create_backup_copy_immediately:
            if backup_level.lower() == 'incremental':
                raise SDKException(
                    'Subclient', '102', 'Backup Copy job is not valid for Transaction Log Backup ')

        request_json = self._sybase_backup_request_json(backup_level.lower(),
                                                        do_not_truncate_log,
                                                        sybase_skip_full_after_logbkp,
                                                        create_backup_copy_immediately,
                                                        backup_copy_type,
                                                        directive_file)

        backup_service = self._commcell_object._services['CREATE_TASK']

        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', backup_service, request_json
        )
        return self._process_backup_response(flag, response)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a Cloud Apps Backupset.

CloudAppsBackupset is the only class defined in this file.

CloudAppsBackupset:  Derived class from Backuset Base class, representing a
                        cloud apps backupset, and to perform operations on that backupset

CloudAppsBackupset:

    __new__()   --  Method to create object based on specific cloud apps instance type


Usage
=====

To add a new Backupset for Cloud Apps agent, please follow these steps:

    1. Add the module for the new instance type under the location:
        **/cvpysdk/backupsets/cloudapps**,
        with the module name **<new instance type>_backupset.py**
        (e.g. "salesforce_backupset.py")

    #. Create a class for your instance type and inherit the CloudAppsBackupset class.

    #. Add the import statement inside the __new__ method.
        **NOTE:** If you add the import statement at the top,
        it'll cause cyclic import, and the call will fail

    #. After adding the import statement:
        - In the **instance_type** dict
            - Add the cloud apps instance type as the key, and the class as its value

"""

from __future__ import unicode_literals

from ..backupset import Backupset


class CloudAppsBackupset(Backupset):
    """Class for representing a Backupset of the Cloud Apps agent."""

    def __new__(cls, instance_object, backupset_name, backupset_id=None):

        from .cloudapps.salesforce_backupset import SalesforceBackupset

        instance_types = {
            3: SalesforceBackupset
        }

        cloud_apps_instance_type = instance_object._properties['cloudAppsInstance']['instanceType']

        if cloud_apps_instance_type in instance_types:
            instance_type = instance_types[cloud_apps_instance_type]
        else:
            instance_type = cls

        return object.__new__(instance_type)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------


"""
ReplicationPairs: Class for a group of replication pairs from the replication monitor

ReplicationPairs Methods:
--------------------------
__init__(commcell_object, **kwargs)     -- Initializes the Replication pairs class with Commcell object and filters
__str__()                               -- Returns the list of all replication pairs as a string
__repr__()                              -- Return the string representation for Replication pairs class
refresh()                               -- Refreshes the list of Replication pairs
_get_replication_pairs()                -- Internal method to fetch replication pairs information from CommServ
**replication_pairs**                   -- Returns the dictionary of replication pair IDs with their information
get(replication_id)                     -- Returns the ReplicationPair class associated to replication ID


ReplicationPair: Class for monitoring a replication pair which exists on periodic monitor

ReplicationPair Attributes:
--------------------------

**vm_pair_id**          -- Returns the live sync VM pair ID

**vm_pair_name**        -- Returns the live sync VM pair name

**replication_guid**    -- Returns the replication guid of the live sync pair

**source_vm**           -- Returns the name of the source virtual machine

**destination_vm**      -- Returns the name of the destination virtual machine

**destination_client**  -- Returns the destination client of the Live sync VM pair

**destination_proxy**   -- Returns the destination proxy of the Live sync VM pair

**destination_instance**-- Returns the destination instance of the Live sync VM pair

**status**              -- Returns the status of the live sync pair

**last_synced_backup_job** -- Returns the last synced backup job ID

**latest_replication_job** -- Returns the latest replication job ID

**last_replication_job**   -- Returns the last replication job ID

**reverse_replication_schedule_id -- Returns the ID of the reverse replication schedule
"""
from ..exception import SDKException
from ..subclients.virtualserver.livesync.vsa_live_sync import LiveSyncVMPair


class ReplicationPairs:
    """Class for a group of replication pairs from the replication monitor"""

    def __init__(self, commcell_object, **kwargs):
        """
        Constructor method for replication monitor with filtering on the basis of kwargs
        Args:
            commcell_object (Commcell): CVPySDK commcell object
            kwargs (dict): The arguments passed for filtering replication pairs
                application_id: The application ID for the replication pairs. eg: 106 for VSA
                instance_id: The instance ID for the source instance associated
                subclient_id: The ID of VM group to find associated pairs for
                schedule_id: The ID of the replication schedule that replication pairs are part of
                failover_group_id: The ID of the failover group that is used for DROrchestration
        """
        self._commcell_object = commcell_object
        self._cypysdk_object = self._commcell_object._cvpysdk_object
        self._services = self._commcell_object._services

        self._filter_args = kwargs

        self._REPLICATION_PAIRS = self._services['GET_REPLICATION_PAIRS']
        if self._filter_args.get('application_id'):
            self._REPLICATION_PAIRS += f'applicationId={self._filter_args.get("application_id")}&'
        if self._filter_args.get('instance_id'):
            self._REPLICATION_PAIRS += f'instanceId={self._filter_args.get("instance_id")}&'
        if self._filter_args.get('subclient_id'):
            self._REPLICATION_PAIRS += f'subclientId={self._filter_args.get("subclient_id")}&'
        if self._filter_args.get('schedule_id'):
            self._REPLICATION_PAIRS += f'taskId={self._filter_args.get("schedule_id")}&'
        if self._filter_args.get('failover_group_id'):
            self._REPLICATION_PAIRS += f'vAppId={self._filter_args.get("failover_group_id")}&'

        self._replication_pairs = {}

        self.refresh()

    def __str__(self):
        """
        Representation string consisting of all Replication VM pairs.
        Returns: (str) table of all replication pairs
        """
        representation_string = ('{:^5}\t{:^20}\t{:^25}\t{:^25}\n\n'
                                 .format('S. No.', 'Replication ID', 'Source Name', 'Destination Name'))

        for index, replication_pair in enumerate(self.replication_pairs):
            replication_pair_dict = self.replication_pairs[replication_pair]
            sub_str = ('{:^5}\t{:^20}\t{:^25}\t{:^25}\n\n'
                       .format(index + 1,
                               replication_pair,
                               replication_pair_dict.get('source_vm'),
                               replication_pair_dict.get('destination_vm')
                               ))
            representation_string += sub_str
        return representation_string.strip()

    def __repr__(self):
        """String representation of the instance of this class."""
        representation_string = f'ReplicationPairs class instance for Filters: "{self._filter_args}"'
        return representation_string

    def refresh(self):
        """Re-populate the replication monitor information"""
        self._get_replication_pairs()

    def _get_replication_pairs(self):
        """
        Gets the list of replication pairs associated to the filters
        Returns: (dict) Dictionary of all replication pairs for the filters
                {
                    "id1": {
                        "source_vm": "sourceVM1",
                        "destination_vm": "sourceVM1DRVM"
                    },
                    "id2": {
                        "source_vm": "sourceVM2",
                        "destination_vm": "sourceVM2DR"
                    }
                }
        Raises:
            SDKException:
                if response is empty
                if response is not success
        """
        flag, response = self._cypysdk_object.make_request('GET', self._REPLICATION_PAIRS)

        if flag:
            if response.json() and 'siteInfo' in response.json():
                replication_pairs = {}
                for site_info in response.json().get('siteInfo', []):
                    replication_pairs[str(site_info.get('replicationId', 0))] = {
                        'source_vm': site_info.get('sourceName', '').lower(),
                        'destination_vm': site_info.get('destinationName', '').lower()
                    }
                return replication_pairs

            raise SDKException('Response', '102')

        response_string = self._commcell_object._update_response_(response.text)
        raise SDKException('Response', '101', response_string)

    @property
    def replication_pairs(self):
        """
        Returns: (dict) Returns the dictionary of all the replication pairs and their info
                {
                    "id1": {
                        "source_vm": "sourceVM1",
                        "destination_vm": "sourceVM1DRVM"
                    },
                    "id2": {
                        "source_vm": "sourceVM2",
                        "destination_vm": "sourceVM2DR"
                    }
                }
        """
        return self._replication_pairs

    def get(self, replication_id: str or int):
        """
        Returns the ReplicationPair object associated to the pair ID
        Args:
            replication_id (str or int)    : The ID of the replication pair
        """
        if replication_id in self.replication_pairs:
            return ReplicationPair(self._commcell_object, str(replication_id))
        raise Exception('ReplicationPairs', '103')


class ReplicationPair(LiveSyncVMPair):
    """Class for replication pair in the replication monitor"""

    def __init__(self, commcell_object, replication_pair_id: int or str):
        """New constructor method which uses pair ID instead of Live sync objects
        Args:
            commcell_object (Commcell): CVPySDK commcell object
            replication_pair_id (int or str): The pair id of the live sync pair
        """
        self._commcell_object = commcell_object
        self._cvpysdk_object = self._commcell_object._cvpysdk_object
        self._services = self._commcell_object._services
        self._update_response_ = self._commcell_object._update_response_

        self._vm_pair_id = str(replication_pair_id)

        self._VM_PAIR = self._services['GET_REPLICATION_PAIR'] % (
            self._vm_pair_id
        )

        self._properties = {}
        self._replication_guid = None
        self._status = None
        self._failover_status = None
        self._source_vm = None
        self._destination_vm = None
        self._destination_client = None
        self._destination_proxy = None
        self._destination_instance = None
        self._last_backup_job = None
        self._latest_replication_job = None

        self._subclient_object = None
        self._subclient_name = None
        self._agent_object = None
        self.live_sync_pair = None

        self.refresh()

        self._populate_live_sync()

    def __repr__(self):
        """String representation of the instance of this class."""
        return f'ReplicationPair class instance for Replication ID: "{self._vm_pair_id}"'

    def __str__(self):
        """String representation of the instance of replication pair"""
        if self._source_vm and self._destination_vm:
            return f'Replication pair: {self._source_vm} -> {self._destination_vm}'
        return f'Replication pair ID: {self._vm_pair_id}'

    def _populate_live_sync(self):
        """ Method used to populate live sync classes"""
        subclient_dict = self._properties.get('parentSubclient', {})
        subtask_dict = self._properties.get('subTask', {})

        client_name = subclient_dict.get('clientName')
        agent_name = self._properties.get('entity', {}).get('appName')
        self._subclient_id = str(subclient_dict.get('subclientId', ''))

        client_object = self._commcell_object.clients.get(client_name)
        self._agent_object = client_object.agents.get(agent_name)

        instance_name = [name for name, instance_id in
                         self._agent_object.instances.all_instances.items()
                         if instance_id == str(subclient_dict.get('instanceId', ''))][0]
        instance_object = self._agent_object.instances.get(instance_name)
        subclient = [(name, subclient_info.get('backupset')) for name, subclient_info in
                     instance_object.subclients.all_subclients.items()
                     if subclient_info.get('id') == self._subclient_id][0]

        backupset_object = instance_object.backupsets.get(subclient[-1])
        self._subclient_object = backupset_object.subclients.get(subclient[0].split('\\')[-1])
        self._subclient_name = self._subclient_object.name

        live_sync = [live_sync_name for live_sync_name, live_sync_dict in
                     self._subclient_object.live_sync.live_sync_pairs.items()
                     if live_sync_dict.get('id') == str(subtask_dict.get('taskId', ''))][0]

        self.live_sync_pair = self._subclient_object.live_sync.get(live_sync)
        self._vm_pair_name = self._source_vm



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a Domino Mailbox Archiver Subclient.

LNDmSubclient is the only class defined in this file.

LNDmSubclient:  Derived class from Subclient Base class.
Represents a domino mailbox archiver subclient, and performs operations on that subclient

LNDmSubclient:

    restore_in_place()                  -- performs an in place restore of the subclient

    restore_out_of_place()              -- performs and out of place restore of the subclient

    backup()                            --  run a backup job for the subclient
"""

from __future__ import absolute_import
from __future__ import unicode_literals

from .lnsubclient import LNSubclient
from ...exception import SDKException


class LNDmSubclient(LNSubclient):
    """Derived class from LNSubclient Base class, representing a LNDM subclient,
        and to perform operations on that subclient."""

    def restore_in_place(
            self,
            paths,
            overwrite=True,
            restore_data_and_acl=True,
            copy_precedence=None,
            from_time=None,
            to_time=None,
            **kwargs):
        """Restores the files/folders specified in the input paths list to the same location.

            Args:
                paths                   (list)  --  list of full paths of files/folders to restore

                overwrite               (bool)  --  unconditional overwrite files during restore

                    default: True

                restore_data_and_acl    (bool)  --  restore data and ACL files

                    default: True

                copy_precedence         (int)   --  copy precedence value of storage policy copy

                    default: None

                from_time           (str)       --  time to retore the contents after

                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                to_time           (str)         --  time to retore the contents before

                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                common_options_dict (dict)          -- dictionary for all the common options
                    options:
                        append                      :   append documents to the database

                            default: False

                        skip                        :   skip if already present

                            default: False

                        unconditionalOverwrite      :   overwrite the documents

                            default: False

                        restoreOnlyStubExists       :   restore only if it is a stub

                            default: False

            Returns:
                object  -   instance of the Job class for this restore job

            Raises:
                SDKException:
                    if paths is not a list

                    if failed to initialize job

                    if response is empty

                    if response is not success

        """
        return super(LNDmSubclient, self).restore_in_place(
            paths,
            overwrite,
            restore_data_and_acl,
            copy_precedence,
            from_time,
            to_time,
            **kwargs)

    def restore_out_of_place(
            self,
            client,
            destination_path,
            paths,
            overwrite=True,
            restore_data_and_acl=True,
            copy_precedence=None,
            from_time=None,
            to_time=None,
            **kwargs):
        """Restores the files/folders specified in the input paths list to the input client,
            at the specified destionation location.

            Args:
                client                (str/object) --  either the name of the client or
                the instance of the Client

                destination_path      (str)        --  full path of the restore location on client

                paths                 (list)       --  list of full paths of
                files/folders to restore

                overwrite             (bool)       --  unconditional overwrite files during restore

                    default: True

                restore_data_and_acl  (bool)       --  restore data and ACL files

                    default: True

                copy_precedence         (int)   --  copy precedence value of storage policy copy

                    default: None

                from_time           (str)       --  time to retore the contents after

                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                to_time           (str)         --  time to retore the contents before

                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                common_options_dict (dict)          -- dictionary for all the common options
                    options:
                        append                      :   append documents to the database

                            default: False

                        skip                        :   skip if already present

                            default: False

                        unconditionalOverwrite      :   overwrite the documents

                            default: False

                        restoreOnlyStubExists       :   restore only if it is a stub

                            default: False

            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:
                    if client is not a string or Client instance

                    if destination_path is not a string

                    if paths is not a list

                    if failed to initialize job

                    if response is empty

                    if response is not success
        """
        return super(LNDmSubclient, self).restore_out_of_place(
            client,
            destination_path,
            paths,
            overwrite,
            restore_data_and_acl,
            copy_precedence,
            from_time,
            to_time,
            **kwargs)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for License operations.

LicenseDetails        : Class for representing license details information

LicenseDetails:
    __init__(Commcell_object)    --  initialise with object of CommCell
    
    _get_detailed_licenses()      --  Gets all types of license details associated to the commcell object
    
    _get_capacity_details()       -- GET request to get capacity licenses property
    
    _get_complete_oi_licenses()    --   GET request to get OI licenses property
    
    _get_virtualization_licenses()    --   GET request to get virtualization licenses property
    
    _get_user_licenses()    --   GET request to get user licenses property
    
    _get_activate_licenses()    --   GET request to get activate licenses property
    
    _get_metallic_licenses()    --   GET request to get metallic licenses property
    
    _get_other_licenses()    --   GET request to get other licenses property
    
    _get_license_details    --   GET request to get detailed license information
    
    
    refresh()    --    Updates License object with the latest configuration
    
    

LicenseDetails Attributes
-------------------------
    commcell_id   --    Returns the CommCell Id in decimal value
    
    commcell_id_hex   --    Returns the hexadecimal value of commcell id
    
    cs_hostname   --    Returns the csHostName Or Address of CommCell
    
    license_ipaddress   --    Returns the license Ip Address 
    
    oem_name   --    Returns the oem_name
    
    license_mode   --    Returns the license mode of license
    
    registration_code   --    Returns the registration code of CommCell
    
    serial_number   --    Returns the serial number of CommCell
    
    expiry_date   --    Returns the expiry date of License
    
    capacity_licenses   --    Returns dictionary with the capacity licenses
    
    complete_oi_licenses   --    Returns dictionary with the complete oi licenses
    
    virtualization_licenses   --    Returns dictionary with the virtualization licenses
    
    user_licenses   --    Returns dictionary with the user licenses
    
    activate_licenses   --    Returns dictionary with the activate licenses
    
    metallic_licenses   --    Returns dictionary with the metallic licenses
    
    other_licenses   --    Returns dictionary with the other licenses

"""
from __future__ import absolute_import
from __future__ import unicode_literals
from .exception import SDKException


class LicenseDetails(object):
    """Class for accessing license details information"""    

    def __init__(self, commcell_object):
        """Initialize object of the LicenseDetails class.

            Args:
                commcell_object (object)  --  instance of the Commcell class

            Returns:
                object - instance of the LicenseDetails class
        """
        self._commcell_object = commcell_object
        self._LICENSE = self._commcell_object._services['LICENSE']
        self._APPLY_LICENSE = self._commcell_object._services['APPLY_LICENSE']
        self._capacity_licenses = None
        self._complete_oi_licenses = None
        self._virtualization_licenses = None
        self._user_licenses = None
        self._activate_licenses = None
        self._metallic_licenses = None
        self._other_licenses = None
        self._get_license_details()

    def _get_detailed_licenses(self):
        """
        Gets all types of license details associated to the commcell object
        """
        self._get_capacity_details()
        self._get_complete_oi_licenses()
        self._get_virtualization_licenses()
        self._get_user_licenses()
        self._get_activate_licenses()
        self._get_metallic_licenses()
        self._get_other_licenses()

    def _get_capacity_details(self):
        """
        Request to get capacity licenses property

        Raises:
            SDKException:

                if response is not success

                if response is empty

        """

        self._CAPACITY_LICENSE = self._commcell_object._services['CAPACITY_LICENSE']
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET', self._CAPACITY_LICENSE)

        if flag:
            if response.json():
                self._capacity_licenses = response.json().get('records',None)
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', response.text)

    def _get_complete_oi_licenses(self):
        """
       GET request to get OI licenses property

        Raises:
            SDKException:

                if response is not success

                if response is empty

        """

        self._OI_LICENSE = self._commcell_object._services['OI_LICENSE']
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET', self._OI_LICENSE)

        if flag:
            if response.json():
                self._complete_oi_licenses = response.json().get('records',None)
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', response.text)

    def _get_virtualization_licenses(self):
        """
       GET request to get virtualization licenses property

        Raises:
            SDKException:

                if response is not success

                if response is empty

        """
        self._VIRTUALIZATION_LICENSE = self._commcell_object._services['VIRTUALIZATION_LICENSE']
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET', self._VIRTUALIZATION_LICENSE)

        if flag:
            if response.json():
                self._virtualization_licenses = response.json().get('records',None)
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', response.text)

    def _get_user_licenses(self):
        """
       GET request to get user licenses property

        Raises:
            SDKException:

                if response is not success

                if response is empty

        """
        self._USER_LICENSE = self._commcell_object._services['USER_LICENSE']
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET', self._USER_LICENSE)

        if flag:
            if response.json():
                self._user_licenses = response.json().get('records',None)
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', response.text)

    def _get_activate_licenses(self):
        """
       GET request to get activate licenses property

        Raises:
            SDKException:

                if response is not success

                if response is empty

        """
        self._ACTIVATE_LICENSE = self._commcell_object._services['ACTIVATE_LICENSE']
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET', self._ACTIVATE_LICENSE)

        if flag:
            if response.json():
                self._activate_licenses = response.json().get('records',None)
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', response.text)

    def _get_metallic_licenses(self):
        """
       GET request to get metallic licenses property

        Raises:
            SDKException:

                if response is not success

                if response is empty

        """
        self._METALLIC_LICENSE = self._commcell_object._services['METALLIC_LICENSE']
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET', self._METALLIC_LICENSE)

        if flag:
            if response.json():
                self._metallic_licenses = response.json().get('records',None)
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', response.text)

    def _get_other_licenses(self):
        """
       GET request to get other licenses property

        Raises:
            SDKException:

                if response is not success

                if response is empty

        """
        self._OTHER_LICENSE = self._commcell_object._services['OTHER_LICENSE']
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET', self._OTHER_LICENSE)

        if flag:
            if response.json():
                self._other_licenses = response.json().get('records',None)
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', response.text)

    def _get_license_details(self):
        """
       GET request to get detailed license information

        Raises:
            SDKException:

                if response is not success

                if response is empty

        """

        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET', self._LICENSE
        )
        if flag:
            if response.json():
                self._commcell_id = response.json()['commcellId']
                self._cs_hostname = response.json()['csHostNameOrAddress']
                self._license_ipaddress = response.json()['licenseIpAddress']
                self._oemname = response.json()['oemName']
                self._regcode = response.json()['regCode']
                self._serialno = response.json()['serialNo']
                self._license_mode = response.json()['licenseMode']
                self._expiry_date = response.json()['expiryDate']
            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101', response.text)
        
    def refresh(self):
        """Updates metrics object with the latest configuration"""
        self._get_license_details()
        self._get_detailed_licenses()

    @property
    def commcell_id(self):
        """Returns the CommCell Id in decimal value"""
        return self._commcell_id

    @property
    def commcell_id_hex(self):
        """Returns the hexadecimal value of commcell id"""
        ccid = self._commcell_id
        if ccid == -1:
            return 'FFFFF'
        return hex(ccid).split('x')[1].upper()

    @property
    def cs_hostname(self):
        """ Returns the csHostName Or Address of CommCell"""
        return self._cs_hostname

    @property
    def license_ipaddress(self):
        """ Returns the license Ip Address """
        return self._license_ipaddress

    @property
    def oem_name(self):
        """Returns the oem_name"""
        return self._oemname

    @property
    def license_mode(self):
        """ Returns the license mode of license"""
        return self._license_mode

    @property
    def registration_code(self):
        """Returns the registration code of CommCell"""
        return self._regcode

    @property
    def serial_number(self):
        """Returns the serial number of CommCell"""
        return self._serialno

    @property
    def expiry_date(self):
        """ Returns the expiry date of License"""
        return self._expiry_date

    @property
    def capacity_licenses(self):
        """Returns dictionary with the capacity licenses"""
        return self._capacity_licenses

    @property
    def complete_oi_licenses(self):
        """Returns dictionary with the complete oi licenses"""
        return self._complete_oi_licenses

    @property
    def virtualization_licenses(self):
        """Returns dictionary with the virtualization licenses"""
        return self._virtualization_licenses

    @property
    def user_licenses(self):
        """Returns dictionary with the user licenses"""
        return self._user_licenses

    @property
    def activate_licenses(self):
        """Returns dictionary with the activate licenses"""
        return self._activate_licenses

    @property
    def metallic_licenses(self):
        """Returns dictionary with the metallic licenses"""
        return self._metallic_licenses

    @property
    def other_licenses(self):
        """Returns dictionary with the other licenses"""
        return self._other_licenses



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a NAS Subclient

NASSubclient is the only class defined in this file.

NASSubclient: Derived class from Subclient Base class, representing a nas subclient,
                        and to perform operations on that subclient

NASSubclient:
    _get_subclient_properties()          --  gets the subclient  related properties of NAS subclient.
    
    _get_subclient_properties_json()     --  gets all the subclient  related properties of NAS subclient.
    
    content()                            --  update the content of the subclient

    filter_content()                    --  update the filter content of the subclient

    content()                           --  update the content of the subclient

    backup()                            --  run a backup job for the subclient
    
    restore_in_place()                  -- run a restore in place for the subclient
    
    restore_out_of_place()                  -- run a restore out of place for the subclient

"""

from __future__ import unicode_literals

from .fssubclient import FileSystemSubclient
from ..exception import SDKException


class NASSubclient(FileSystemSubclient):
    """Derived class from Subclient Base class, representing a nas subclient,
        and to perform operations on that subclient."""
    
   
    def backup(
            self,
            backup_level="Incremental",
            incremental_backup=False,
            incremental_level='BEFORE_SYNTH',
            on_demand_input=None,
            snap_name=None,
            backup_offline_data=False,
            volume_based_backup=False):
        """Runs a backup job for the subclient of the level specified.

            Args:
                backup_level        (str)   --  level of backup the user wish to run
                        Full / Incremental / Differential / Synthetic_full
                    default: Incremental

                incremental_backup  (bool)  --  run incremental backup
                        only applicable in case of Synthetic_full backup
                    default: False

                incremental_level   (str)   --  run incremental backup before/after synthetic full
                        BEFORE_SYNTH / AFTER_SYNTH

                        only applicable in case of Synthetic_full backup
                    default: BEFORE_SYNTH

                on_demand_input     (str)   --  input file location for on demand backupset
                    default: None

                snap_name   (str)   --  input for snap_name

                backup_offline_data   (bool)  --  input for advanced NAS backup option backupOfflineData

                volume_based_backup   (bool)  --  input for advanced NAS backup option volumeBasedBackup

            Returns:
                object - instance of the Job class for this backup job

            Raises:
                SDKException:
                    if backup level specified is not correct

                    if response is empty

                    if response is not success
        """
        request_json = self._backup_json(
            backup_level,
            incremental_backup,
            incremental_level,
            on_demand_input)

        nas_options = {
            "nasOptions": {
                "backupFromSnap": snap_name if snap_name is not None else "",
                "backupOfflineData": True if backup_offline_data else False,
                "backupQuotas": True,
                "backupFromSnapshot": True,
                "backupFromSnapshotYes": True,
                "replicationVolumeId": 0,
                "volumeBasedBackup": True if volume_based_backup else False
            }
        }

        request_json["taskInfo"]["subTasks"][0]["options"]["backupOpts"].update(
            nas_options
        )

        bakup_service = self._commcell_object._services['CREATE_TASK']

        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', bakup_service, request_json
        )

        return self._process_backup_response(flag, response)

    def restore_in_place(
            self,
            paths,
            overwrite=None,
            restore_data_and_acl=True,
            copy_precedence=None,
            from_time=None,
            to_time=None,
            fs_options=None,
            schedule_pattern=None,
            proxy_client=None,
            advanced_options=None,
            synth_restore=False,
            DAR=None,
            noRecursive=None
            ):
        """Runs a restore job for the subclient .

            Args:
                            paths                   (list)  --  list of full paths of files/folders to restore

                overwrite               (bool)  --  unconditional overwrite files during restore
                    default: None

                restore_data_and_acl    (bool)  --  restore data and ACL files
                    default: True

                copy_precedence         (int)   --  copy precedence value of storage policy copy
                    default: None

                from_time           (str)       --  time to retore the contents after
                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                to_time           (str)         --  time to retore the contents before
                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                fs_options      (dict)          -- dictionary that includes all advanced options
                    options:
                        all_versions        : if set to True restores all the versions of the
                                                specified file
                        versions            : list of version numbers to be backed up
                        validate_only       : To validate data backed up for restore


                schedule_pattern (dict) -- scheduling options to be included for the task

                        Please refer schedules.schedulePattern.createSchedule()
                                                                    doc for the types of Jsons

                proxy_client    (str)          -- Proxy client used during FS under NAS operations

                advanced_options    (dict)  -- Advanced restore options
                
                synth_restore (bool)     -- Advance NAS restore option SynthRestore

                DAR (bool)     -- Advance NAS restore option DAR

                noRecursive (bool)     -- Advance NAS restore option Recursive

                    Options:

                        job_description (str)   --  Restore job description

                        timezone        (str)   --  Timezone to be used for restore

                            **Note** make use of TIMEZONES dict in constants.py to pass timezone

            Returns:
                object - instance of the Job class for this restore job if its an immediate Job
                         instance of the Schedule class for this restore job if its a scheduled Job

            Raises:
                SDKException:
                    if paths is not a list

                    if failed to initialize job

                    if response is empty

                    if response is not success
                
        """
        self._backupset_object._instance_object._restore_association = self._subClientEntity

        request_json = self._restore_json(
            paths=paths,
            overwrite=overwrite,
            restore_data_and_acl=restore_data_and_acl,
            copy_precedence=copy_precedence,
            from_time=from_time,
            to_time=to_time,
            fs_options=fs_options,
            schedule_pattern=None,
            proxy_client=None,
            advanced_options=None
        )

        request_json['taskInfo']['subTasks'][0]['options']['restoreOptions']['commonOptions']['unconditionalOverwrite'] = True
        request_json['taskInfo']['subTasks'][0]['options']['restoreOptions']['commonOptions']['stripLevel'] = 0
        request_json['taskInfo']['subTasks'][0]['options']['restoreOptions']['commonOptions']['preserveLevel'] = 1
        request_json['taskInfo']['subTasks'][0]['options']['restoreOptions']['commonOptions']['stripLevelType'] = 0

        if fs_options:
            request_json['taskInfo']['subTasks'][0]['options']['restoreOptions']['browseOption']['liveBrowse'] = True

        nas_option = {
            "nasOption": {
                "synthRestore": 1 if synth_restore is True else 2,
                "useDirectAccess": 0 if DAR is True else 1,
                "noRecursive": False if noRecursive is False else True,
                "overwrite": True if overwrite is True else False
            }
        }

        request_json["taskInfo"]["subTasks"][0]["options"]["restoreOptions"].update(nas_option)

        return self._process_restore_response(request_json)
        
    def restore_out_of_place(
            self,
            client,
            destination_path,
            paths,
            overwrite=None,
            restore_data_and_acl=True,
            copy_precedence=None,
            from_time=None,
            to_time=None,
            fs_options=None,
            schedule_pattern=None,
            proxy_client=None,
            advanced_options=None,
            synth_restore=False,
            DAR=None,
            noRecursive=None
            ):
        """Runs a restore job for the subclient .

            Args:
                            paths                   (list)  --  list of full paths of files/folders to restore

                overwrite               (bool)  --  unconditional overwrite files during restore
                    default: True

                restore_data_and_acl    (bool)  --  restore data and ACL files
                    default: True

                copy_precedence         (int)   --  copy precedence value of storage policy copy
                    default: None

                from_time           (str)       --  time to retore the contents after
                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                to_time           (str)         --  time to retore the contents before
                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                fs_options      (dict)          -- dictionary that includes all advanced options
                    options:
                        all_versions        : if set to True restores all the versions of the
                                                specified file
                        versions            : list of version numbers to be backed up
                        validate_only       : To validate data backed up for restore


                schedule_pattern (dict) -- scheduling options to be included for the task

                        Please refer schedules.schedulePattern.createSchedule()
                                                                    doc for the types of Jsons

                proxy_client    (str)          -- Proxy client used during FS under NAS operations

                advanced_options    (dict)  -- Advanced restore options
                
                synth_restore (bool)     -- Advance NAS restore option SynthRestore

                DAR (bool)     -- Advance NAS restore option DAR

                noRecursive (bool)     -- Advance NAS restore option Recursive

                    Options:

                        job_description (str)   --  Restore job description

                        timezone        (str)   --  Timezone to be used for restore

                            **Note** make use of TIMEZONES dict in constants.py to pass timezone

            Returns:
                object - instance of the Job class for this restore job if its an immediate Job
                         instance of the Schedule class for this restore job if its a scheduled Job

            Raises:
                SDKException:
                    if paths is not a list

                    if failed to initialize job

                    if response is empty

                    if response is not success
                
        """
        self._backupset_object._instance_object._restore_association = self._subClientEntity

        request_json = self._restore_json(
            client=client,
            destPath=destination_path,
            paths=paths,
            overwrite=overwrite,
            restore_data_and_acl=restore_data_and_acl,
            copy_precedence=copy_precedence,
            from_time=from_time,
            to_time=to_time,
            fs_options=fs_options,
            schedule_pattern=None,
            proxy_client=None,
            advanced_options=None
        )

        destination_options = request_json['taskInfo']['subTasks'][0]['options']['restoreOptions'].get('destination', {})
        destination_options['destPath'] = destination_options.get('destPath', [''])
        request_json['taskInfo']['subTasks'][0]['options']['restoreOptions']['destination']['destPath'][0] = destination_path
        request_json['taskInfo']['subTasks'][0]['options']['restoreOptions']['destination']['inPlace'] = False

        request_json['taskInfo']['subTasks'][0]['options']['restoreOptions']['commonOptions']['unconditionalOverwrite'] = True
        request_json['taskInfo']['subTasks'][0]['options']['restoreOptions']['commonOptions']['stripLevel'] = 2
        request_json['taskInfo']['subTasks'][0]['options']['restoreOptions']['commonOptions']['preserveLevel'] = 1
        request_json['taskInfo']['subTasks'][0]['options']['restoreOptions']['commonOptions']['stripLevelType'] = 0

        if fs_options:
            request_json['taskInfo']['subTasks'][0]['options']['restoreOptions']['browseOption']['liveBrowse'] = True

        nas_option = {
            "nasOption": {
                "synthRestore": 1 if synth_restore is True else 2,
                "useDirectAccess": 0 if DAR is True else 1,
                "noRecursive": False if noRecursive is False else True,
                "overwrite": True if overwrite is True else False
            }
        }

        request_json['taskInfo']['subTasks'][0]['options']['restoreOptions'].update(nas_option)

        return self._process_restore_response(request_json)


# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Main file for performing operations on file storage optimization(FSO) app under Activate.

'FsoTypes', 'FsoServers' , 'FsoServer', 'FsoServerGroups', 'FsoServerGroup' are 5 classes defined in this file

FsoTypes:  Class to represent different FSO types(Server/ServerGroup/Project)

FsoServers: Class to represent all FSO servers in the commcell

FsoServer:  Class to represent single FSO server in the commcell

FsoServerGroups:  Class to represent all FSO server groups in the commcell

FsoServerGroup:   Class to represent single FSO server group in the commcell

FsoServers:

    __init__()                          --  initialise object of the FsoServers class

     _response_not_success()            --  parses through the exception response, and raises SDKException

     _get_all_fso_servers()             --  gets all the fso servers from the commcell

     refresh()                          --  refresh the FSO Servers associated with the commcell

     has_server()                       --  checks whether given server name exists in FSO or not

     add_file_server()                  --  adds file server to the FSO

     get()                              --  returns the FsoServer object for given server name

FsoServer:

    __init__()                          --  initialise object of the FsoServer class

     _response_not_success()            --  parses through the exception response, and raises SDKException

     _get_server_properties()           --  gets FSO server details from the commcell

     _get_schedule_object()             --  returns the schedule object for associated schedule

     refresh()                          --  refresh the FSO Server details

     start_collection()                 --  starts collection job on all data sources in this server

     share()                            --  shares server with given user name or group name

     search()                           --  returns the search response containing document details

     add_schedule()                     --  creates schedule for this fso server

     delete_schedule()                  --  deletes schedule for this fso server

FsoServer Attributes
---------------------

    **server_id**           --  returns the client id of the server

    **server_details**      --  returns the server details

    **data_sources**        --  returns the EdiscoveryDataSources object

    **data_sources_name**   --  returns the list of data sources display name associated with this server

    **total_data_sources**  --  returns the total number of data sources associated with this server

    **total_doc_count**     --  returns the total document count from all data sources associated with this server

    **schedule**            --  returns the schedule object for associated schedule with this server

FsoServerGroups:

    __init__()                          --  initialise object of the FsoServerGroups class

     _response_not_success()            --  parses through the exception response, and raises SDKException

     _get_all_fso_server_groups()       --  gets all the fso server groups from the commcell

     refresh()                          --  refresh the FSO Server groups associated with the commcell

     has()                              --  checks whether given server group name exists in FSO or not

     get()                              --  returns object of FsoServerGroup class

     add_server_group()                 --  adds server group to FSO

FsoServerGroup:

    __init__()                          --  initialise object of the FsoServerGroup class

     _response_not_success()            --  parses through the exception response, and raises SDKException

     _get_server_group_properties()     --  gets FSO server group details from the commcell

     refresh()                          --  refresh the FSO Server group details

     has_server()                       --  returns whether server name exists as part of server group or not

     get()                              --  returns the object of FsoServer class

     start_collection()                 --  starts collection job on all servers associated with this server group

     search()                           --  returns the search response containing document details

FsoServerGroup Attributes:
---------------------------

    **server_group_id**     --  returns the server group id

    **server_group_props**  --  returns the properties of server group

    **server_list**         --  returns the list of servers associated with this server group

    **total_documents**     --  returns the total crawled document count for this server group

"""
import copy
import time
from enum import Enum

from ..schedules import Schedules

from ..activateapps.constants import EdiscoveryConstants

from ..activateapps.ediscovery_utils import EdiscoveryClients, EdiscoveryClientOperations, EdiscoveryDataSources

from ..exception import SDKException


class FsoTypes(Enum):
    """Class to represent different FSO types(Server/ServerGroup/Project)"""
    SERVERS = 0
    SERVER_GROUPS = 1
    PROJECTS = 2


class FsoServers():
    """Class for representing all FSO servers in the commcell."""

    def __init__(self, commcell_object):
        """Initializes an instance of the FsoServers class.

            Args:
                commcell_object     (object)    --  instance of the commcell class

            Returns:
                object  -   instance of the FsoServers class

        """
        self._commcell_object = commcell_object
        self._update_response_ = commcell_object._update_response_
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._fso_servers = None
        self._ediscovery_clients_obj = EdiscoveryClients(self._commcell_object, self)
        self._ediscovery_ds_obj = EdiscoveryDataSources(self._commcell_object, self)
        self.refresh()

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._update_response_(response.text))

    def get(self, server_name):
        """returns the FsoServer object for given server name

                Args:

                    server_name         (str)       --  Name of the server

                Returns:

                    obj --  Instance of FsoServer Class

                Raises:

                    SDKException:

                            if failed to find server in FSO App

                            if input is not valid

        """
        if not isinstance(server_name, str):
            raise SDKException('FileStorageOptimization', '101')
        if not self.has_server(server_name):
            raise SDKException('FileStorageOptimization', '103')
        server_id = self._fso_servers[server_name.lower()]['clientEntity']['clientId']
        return FsoServer(commcell_object=self._commcell_object, server_name=server_name, server_id=server_id)

    def _get_all_fso_servers(self):
        """Returns all the FSO servers found in the commcell

                Args:

                    None

                Returns:

                    dict        --  Containing FSO server details

                Raises;

                    SDKException:

                            if failed to get FSO servers details

                            if response is empty

                            if response is not success
        """
        return self._ediscovery_clients_obj.get_ediscovery_clients()

    def refresh(self):
        """Refresh the FSO Servers associated with the commcell."""
        self._fso_servers = self._get_all_fso_servers()

    def add_file_server(self, server_name, data_source_name, inventory_name, plan_name,
                        source_type=EdiscoveryConstants.SourceType.BACKUP, **kwargs):
        """Adds file system FSO server

                Args:

                    server_name         (str)       --  Server name which needs to be added

                    data_source_name    (str)       --  Name for data source

                    inventory_name      (str)       --  Inventory name which needs to be associated

                    plan_name           (str)       --  Plan name which needs to be associated with this data source

                    source_type         (enum)      --  Source type for crawl (Live source or Backedup)
                                                                Refer EdiscoveryConstants.SourceType

                Kwargs Arguments:

                    scan_type           (str)       --  Specifies scan type when source type is for backed up data
                                                                Supported values : quick | full

                    crawl_path          (list)      --  File path which needs to be crawl if source type is Live source

                    access_node         (str)       --  server name which needs to be used as access node
                                                            in case if server to be added is not a commvault client

                    country_name        (str)       --  country name where server is located (default : USA)

                    country_code        (str)       --  Country code (ISO 3166 2-letter code)

                    user_name           (str)       --  User name who has access to UNC path

                    password            (str)       --  base64 encoded password to access unc path

                    enable_monitoring   (str)       --  specifies whether to enable file monitoring or not for this


                Returns:

                    obj     --  Instance of FSOServer class

                Raises:

                      SDKException:

                            if plan/inventory/index server doesn't exists

                            if failed to add FSO server data source
        """
        self._ediscovery_ds_obj.add_fs_data_source(
            server_name=server_name,
            data_source_name=data_source_name,
            inventory_name=inventory_name,
            plan_name=plan_name,
            source_type=source_type,
            **kwargs)
        is_commvault_client = self._commcell_object.clients.has_client(server_name)
        server_id = 0
        if not is_commvault_client:
            all_clients = self._commcell_object.clients.all_clients
            for client_name, client_details in all_clients.items():
                if client_name.lower().startswith(f"{data_source_name.lower()}_"):
                    server_id = client_details['id']
                    break
        else:
            server_id = self._commcell_object.clients.get(server_name).client_id
        return FsoServer(commcell_object=self._commcell_object, server_name=server_name,
                         server_id=server_id)

    def has_server(self, server_name):
        """Checks if a server exists in the commcell with the input name for FSO or not

            Args:
                server_name (str)  --  name of the server

            Returns:
                bool - boolean output whether the FSO Server exists in the commcell or not

            Raises:
                SDKException:
                    if type of the server name argument is not string

        """
        if not isinstance(server_name, str):
            raise SDKException('FileStorageOptimization', '101')
        return self._fso_servers and server_name.lower() in self._fso_servers


class FsoServer():
    """Class to represent single FSO Server in the commcell"""

    def __init__(self, commcell_object, server_name, server_id=None):
        """Initializes an instance of the FsoServer class.

            Args:
                commcell_object     (object)    --  instance of the commcell class

                server_name         (str)       --  Name of the server

                server_id           (int)       --  server client id

            Returns:
                object  -   instance of the FsoServer class

        """
        self._commcell_object = commcell_object
        self._update_response_ = commcell_object._update_response_
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._server_name = server_name
        self._server_id = None
        self._server_props = None
        self._schedule_obj = None
        self._CREATE_POLICY = self._services['CREATE_UPDATE_SCHEDULE_POLICY']
        if server_id:
            self._server_id = server_id
        else:
            self._server_id = self._commcell_object.activate.file_storage_optimization().get(server_name).server_id
        self._ediscovery_client_ops = EdiscoveryClientOperations(self._commcell_object, self)
        self._ediscovery_data_srcs_obj = EdiscoveryDataSources(self._commcell_object, self)
        self.refresh()

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._update_response_(response.text))

    def start_collection(self):
        """Starts collection job on all data sources associated with this server

                Args:

                    None

                Return:

                    list    --  List of jobid's

                Raises:

                    SDKException:

                            if failed to start collection job

        """
        request_json = copy.deepcopy(EdiscoveryConstants.START_CRAWL_SERVER_REQUEST_JSON)
        request_json['taskInfo']['associations'][0]['clientId'] = self._server_id
        request_json['taskInfo']['task'][
            'taskName'] = f"Cvpysdk_FSO_server_Crawl_{self._server_name}_{int(time.time())}"
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', self._CREATE_POLICY, request_json
        )
        output = []
        if flag:
            if response.json() and 'jobIds' in response.json():
                for node in response.json()['jobIds']:
                    output.append(node)
                return output
            raise SDKException('FileStorageOptimization', '105')
        self._response_not_success(response)

    def _get_schedule_object(self):
        """returns the schedule object for associated schedule for this fso server

            Args:
                None

            Returns:

                obj --  Instance of Schedule class

                None -- if no schedule exists

            Raises:

                SDKException:

                        if failed to find schedule details associated with this server
        """
        scd_obj = Schedules(self)
        if scd_obj.has_schedule():
            return scd_obj.get()
        return None

    def _get_server_properties(self):
        """gets FSO server details from the commcell

                Args:

                    None

                Returns:

                    dict    --  Containing FSO Server details

                Raises:

                     Raises;

                        SDKException:

                            if failed to get server details

        """
        self._ediscovery_data_srcs_obj.refresh()  # do refresh before fetching so that doc count comes up fine
        return self._ediscovery_data_srcs_obj.ediscovery_client_props

    def search(self, criteria=None, attr_list=None, params=None):
        """do searches on data source and returns document details

            Args:

                criteria        (str)      --  containing criteria for query
                                                    (Default : None - returns all docs)

                                                    Example :

                                                        Size:[10 TO 1024]
                                                        FileName:09_23*

                attr_list       (set)      --  Column names to be returned in results.
                                                     Acts as 'fl' in query

                params          (dict)     --  Any other params which needs to be passed
                                                   Example : { "start" : "0" }

            Returns:

                int,list(dict),dict    --  Containing document count, document details & facet details(if any)

            Raises:

                SDKException:

                        if failed to perform search

        """
        return self._ediscovery_client_ops.search(criteria=criteria, attr_list=attr_list, params=params)

    def add_schedule(self, schedule_name, pattern_json):
        """Creates the schedule and associate it with server

                        Args:

                            schedule_name       (str)       --  Schedule name

                            pattern_json        (dict)      --  Schedule pattern dict
                                                                (Refer to Create_schedule_pattern in schedule.py)

                        Raises:

                              SDKException:

                                    if input is not valid

                                    if failed to create schedule

        """
        self._ediscovery_client_ops.schedule(schedule_name=schedule_name, pattern_json=pattern_json)
        self.refresh()

    def delete_schedule(self):
        """Deletes the schedule associated with server

                        Args:

                            None

                        Raises:

                              SDKException:

                                    if failed to Delete schedule

        """
        if not self._schedule_obj:
            raise SDKException('FileStorageOptimization', '102', "No schedule is associated to this FSO Server")
        Schedules(self).delete()
        self.refresh()

    def share(self, user_or_group_name, allow_edit_permission=False, is_user=True, ops_type=1):
        """Shares Fso server with given user or user group in commcell

                Args:

                    user_or_group_name      (str)       --  Name of user or group

                    is_user                 (bool)      --  Denotes whether this is user or group name
                                                                default : True(User)

                    allow_edit_permission   (bool)      --  whether to give edit permission or not to user or group

                    ops_type                (int)       --  Operation type

                                                            Default : 1 (Add)

                                                            Supported : 1 (Add)
                                                                        3 (Delete)

                Returns:

                    None

                Raises:

                    SDKException:

                            if unable to update security associations

                            if response is empty or not success
        """
        return self._ediscovery_client_ops.share(
            user_or_group_name=user_or_group_name,
            allow_edit_permission=allow_edit_permission,
            is_user=is_user,
            ops_type=ops_type)

    def refresh(self):
        """Refresh the FSO Server details"""
        self._server_props = self._get_server_properties()
        self._schedule_obj = self._get_schedule_object()

    @property
    def schedule(self):
        """returns the schedule object for associated schedule

                Returns:

                    obj     --  Instance of Schedule Class if schedule exists

                    None    --  If no schedule exists

        """
        return self._schedule_obj

    @property
    def server_id(self):
        """returns the server id

            Returns:

                int --  Server id

        """
        return self._server_id

    @property
    def server_details(self):
        """returns the server details

            Returns:

                dict --  Server details

        """
        return self._server_props

    @property
    def data_sources_name(self):
        """returns the associated data sources to this FSO server

            Returns:

                list --  names of data sources

        """
        return self._ediscovery_data_srcs_obj.data_sources

    @property
    def data_sources(self):
        """returns the EdiscoveryDataSources object associated to this server

            Returns:

                obj --  Instance of EdiscoveryDataSources Object

        """
        return self._ediscovery_data_srcs_obj

    @property
    def total_data_sources(self):
        """returns the total number of data sources associated with this server

            Returns:

                int --  total number of data sources

        """
        return len(self._ediscovery_data_srcs_obj.data_sources)

    @property
    def total_doc_count(self):
        """returns the total document count of all data sources for this server

            Returns:

                int --  Total crawled document count

        """
        return self._ediscovery_data_srcs_obj.total_documents


class FsoServerGroups():
    """Class for representing all FSO server groups in the commcell."""

    def __init__(self, commcell_object):
        """Initializes an instance of the FsoServerGroups class.

            Args:
                commcell_object     (object)    --  instance of the commcell class

            Returns:
                object  -   instance of the FsoServerGroups class

        """
        self._commcell_object = commcell_object
        self._update_response_ = commcell_object._update_response_
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._fso_server_groups = None
        self._ediscovery_clients_obj = EdiscoveryClients(self._commcell_object, self)
        self._ediscovery_ds_obj = EdiscoveryDataSources(self._commcell_object, self)
        self.refresh()

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._update_response_(response.text))

    def _get_all_fso_server_groups(self):
        """Returns all the FSO server groups found in the commcell

                Args:

                    None

                Returns:

                    dict        --  Containing FSO server group details

                Raises;

                    SDKException:

                            if failed to get FSO server group details

                            if response is empty

                            if response is not success
        """
        return self._ediscovery_clients_obj.get_ediscovery_clients()

    def refresh(self):
        """Refresh the FSO Server groups associated with the commcell."""
        self._fso_server_groups = self._get_all_fso_server_groups()

    def add_server_group(self, server_group_name, inventory_name, plan_name, **kwargs):
        """adds server group to FSO

                Args:

                    server_group_name       (str)       --      Server group name

                    inventory_name          (str)       --  Inventory name which needs to be associated

                    plan_name               (str)       --  Plan name which needs to be associated with this data source

                 Kwargs Arguments:

                    country_name        (str)       --  country name where server is located (default: USA)

                    country_code        (str)       --  Country code (ISO 3166 2-letter code)

                Returns:

                    obj     --  Instance of FSOServerGroup class

                Raises:

                      SDKException:

                            if plan/inventory/index server doesn't exists

                            if failed to add FSO server group

        """
        self._ediscovery_ds_obj.add_fs_data_source(
            server_name=server_group_name,
            data_source_name=server_group_name,
            inventory_name=inventory_name,
            plan_name=plan_name,
            **kwargs)
        return FsoServerGroup(
            self._commcell_object,
            server_group_name,
            server_id=self._commcell_object.client_groups.get(server_group_name).clientgroup_id)

    def has(self, server_group_name):
        """Checks if a server group exists in the commcell with the input name for FSO or not

            Args:
                server_group_name (str)  --  name of the server group

            Returns:
                bool - boolean output whether the FSO Server group exists in the commcell or not

            Raises:
                SDKException:
                    if type of the server group name argument is not string

        """
        if not isinstance(server_group_name, str):
            raise SDKException('FileStorageOptimization', '101')
        return self._fso_server_groups and server_group_name.lower() in self._fso_server_groups

    def get(self, server_grp_name):
        """returns the FsoServerGroup object for given server group name

                Args:

                    server_grp_name         (str)       --  Name of the server group

                Returns:

                    obj --  Instance of FsoServerGroup Class

                Raises:

                    SDKException:

                            if failed to find server group in FSO App

                            if input is not valid

        """
        if not isinstance(server_grp_name, str):
            raise SDKException('FileStorageOptimization', '101')
        if not self.has(server_grp_name):
            raise SDKException('FileStorageOptimization', '106')
        server_id = self._fso_server_groups[server_grp_name.lower()]['clientEntity']['clientId']
        return FsoServerGroup(
            commcell_object=self._commcell_object,
            server_group_name=server_grp_name,
            server_id=server_id)


class FsoServerGroup():
    """Class to represent single FSO Server group in the commcell"""

    def __init__(self, commcell_object, server_group_name, server_id=None):
        """Initializes an instance of the FsoServerGroup class.

            Args:
                commcell_object     (object)    --  instance of the commcell class

            Returns:
                object  -   instance of the FsoServerGroup class

        """
        self._commcell_object = commcell_object
        self._update_response_ = commcell_object._update_response_
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._server_group_name = server_group_name
        self._server_id = None
        self._server_grp_props = None
        self._server_names = []
        self._total_doc = 0
        self._CREATE_POLICY = self._services['CREATE_UPDATE_SCHEDULE_POLICY']
        if server_id:
            self._server_id = server_id
        else:
            self._server_id = self._commcell_object.activate.file_storage_optimization(
                FsoTypes.SERVER_GROUPS).get(server_group_name).server_group_id
        self._ediscovery_clients_obj = EdiscoveryClients(self._commcell_object, self)
        self._ediscovery_client_ops = EdiscoveryClientOperations(self._commcell_object, self)
        self.refresh()

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._update_response_(response.text))

    def _get_server_group_properties(self):
        """gets FSO server group details from the commcell

                Args:

                    None

                Returns:

                    dict    --  Containing FSO Server group details

                Raises:

                     Raises;

                        SDKException:

                            if failed to get server group details

        """
        client_resp = self._ediscovery_clients_obj.get_ediscovery_clients()
        grp_resp = self._ediscovery_clients_obj.get_ediscovery_client_group_details()
        if 'nodeList' in grp_resp:
            grp_resp = grp_resp['nodeList'][0]
            if 'childs' in grp_resp and 'customProperties' in grp_resp['childs'][0]:
                name_value_dict = grp_resp['childs'][0]['customProperties']['nameValues']
                for prop in name_value_dict:
                    prop_name = prop.get('name')
                    if prop_name == EdiscoveryConstants.FIELD_DOCUMENT_COUNT:
                        self._total_doc = int(prop.get('value'))
                        break
        self._server_names = []
        for key, value in client_resp.items():
            self._server_names.append(key)
        return client_resp

    def has_server(self, server_name):
        """Checks if a server exists in the FSO Server group with the input name or not

            Args:
                server_name (str)  --  name of the server

            Returns:
                bool - boolean output whether the FSO Server exists in the server group or not

            Raises:
                SDKException:
                    if type of the server name argument is not string

        """
        if not isinstance(server_name, str):
            raise SDKException('FileStorageOptimization', '101')
        return self._server_grp_props and server_name.lower() in self._server_grp_props

    def start_collection(self):
        """Starts collection job on all servers associated with this server group

                Args:

                    None

                Return:

                    list    --  List of jobid's

                Raises:

                    SDKException:

                            if failed to start collection job

        """
        request_json = copy.deepcopy(EdiscoveryConstants.START_CRAWL_SERVER_REQUEST_JSON)
        # delete clientid key and add client group level key and entity type
        del request_json['taskInfo']['associations'][0]['clientId']
        request_json['taskInfo']['associations'][0]['clientGroupId'] = self._server_id
        request_json['taskInfo']['associations'][0]['_type_'] = 28  # server group level job collection
        request_json['taskInfo']['task'][
            'taskName'] = f"Cvpysdk_FSO_server_Crawl_{self._server_group_name}_{int(time.time())}"
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', self._CREATE_POLICY, request_json
        )
        output = []
        if flag:
            if response.json() and 'jobIds' in response.json():
                for node in response.json()['jobIds']:
                    output.append(node)
                return output
            raise SDKException('FileStorageOptimization', '107')
        self._response_not_success(response)

    def search(self, criteria=None, attr_list=None, params=None):
        """do searches on client group data and returns document details

            Args:

                criteria        (str)      --  containing criteria for query
                                                    (Default : None - returns all docs)

                                                    Example :

                                                        Size:[10 TO 1024]
                                                        FileName:09_23*

                attr_list       (set)      --  Column names to be returned in results.
                                                     Acts as 'fl' in query

                params          (dict)     --  Any other params which needs to be passed
                                                   Example : { "start" : "0" }

            Returns:

                int,list(dict),dict    --  Containing document count, document details & facet details(if any)

            Raises:

                SDKException:

                        if failed to perform search

        """
        return self._ediscovery_client_ops.search(criteria=criteria, attr_list=attr_list, params=params)

    def get(self, server_name):
        """returns the FsoServer object for given server name

                Args:

                    server_name         (str)       --  Name of the server

                Returns:

                    obj --  Instance of FsoServer Class

                Raises:

                    SDKException:

                            if failed to find server in FSO server group

                            if input is not valid

        """
        if not isinstance(server_name, str):
            raise SDKException('FileStorageOptimization', '101')
        if not self.has_server(server_name):
            raise SDKException('FileStorageOptimization', '103')
        server_id = self._server_grp_props[server_name.lower()]['clientEntity']['clientId']
        return FsoServer(commcell_object=self._commcell_object, server_name=server_name, server_id=server_id)

    def refresh(self):
        """Refresh the FSO Server group details"""
        self._server_grp_props = self._get_server_group_properties()

    @property
    def server_group_id(self):
        """returns the client group id

            Returns:

                int --  Server group id

        """
        return self._server_id

    @property
    def server_group_props(self):
        """returns the server group properties

            Returns:

                dict --  Server group details

        """
        return self._server_grp_props

    @property
    def server_list(self):
        """returns the list of server which is part of this server group

            Returns:

                list --  Server names

        """
        return self._server_names

    @property
    def total_documents(self):
        """returns the total documents count for this server group

            Returns:

                int --  Total document count for this server group

        """
        return self._total_doc



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Helper file to manage security associations on this commcell

SecurityAssociation is the only class defined in this file

SecurityAssociation:
    __init__()                  --  initializes security class object

    __str__()                   --  returns all the users associated with the commcell

    __repr__()                  --  returns the string for the instance of the User class

    _security_association_json()--  generates security association blob with all
                                    user-entity-role association

    fetch_security_association()--  fetches security associations from entity

    _get_security_roles()       --  gets the list of all the security roles applicable
                                        on this commcell

    _add_security_association() --  adds the security association with client or clientgroup

    has_role()                  --  checks if specified role exists on commcell


"""

from ..exception import SDKException


class SecurityAssociation(object):
    """Class for managing the security associations roles on the commcell"""

    def __init__(self, commcell_object, class_object):
        """Initializes the security associations object

            Args:
                commcell_object     (object)     --     instance of the Commcell class

                class_object         (object)     --    instance of the class on which we want to
                                                            manage security operations
        """
        self._commcell_object = commcell_object

        from ..client import Client
        if isinstance(class_object, Client):
            self._entity_list = {
                "entity": [{
                    "clientId": int(class_object.client_id),
                    "_type_": 3
                }]
            }

        from ..storage_pool import StoragePool
        if isinstance(class_object, StoragePool):
            self._entity_list = {
                "entity": [{
                    "storagePolicyId": int(class_object.storage_pool_id),
                    "_type_": 17
                }]
            }
            
        from ..plan import Plan
        if isinstance(class_object, Plan):
            self._entity_list = {
                "entity": [{
                    "planId": int(class_object.plan_id),
                    "_type_": 158
                }]
            }

        self._roles = self._get_security_roles()

    def __str__(self):
        """Representation string consisting of all available security roles on this commcell.

            Returns:
                str - string of all the available security roles on this commcell
        """
        representation_string = '{:^5}\t{:^20}\n\n'.format('S. No.', 'Roles')

        for index, role in enumerate(self._roles):
            sub_str = '{:^5}\t{:20}\n'.format(index + 1, role)
            representation_string += sub_str

        return representation_string.strip()

    def __repr__(self):
        """Representation string for the instance of the Security class."""
        return "Security class instance for Commcell: '{0}'".format(
            self._commcell_object.commserv_name
        )

    @staticmethod
    def _security_association_json(entity_dictionary):
        """handles three way associations (role-user-entities)

            Args:
                entity_dictionary   --      combination of entity_type, entity names
                                            and role
                e.g.: entity_dict={
                                'assoc1':
                                    {
                                        'entity_type':['entity_name'],
                                        'entity_type':['entity_name', 'entity_name'],
                                        'role': ['role1']
                                    },
                                'assoc2':
                                    {
                                        'mediaAgentName': ['networktestcs', 'standbycs'],
                                        'clientName': ['Linux1'],
                                        'role': ['New1']
                                        }
                                    }
                entity_type         --      key for the entity present in dictionary
                                            on which user will have access

                entity_name         --      Value of the key

                role                --      role will remain role in dictionary
                e.g.: {"clientName":"Linux1"}
                entity_type:    clientName, mediaAgentName, libraryName, userName,
                                userGroupName, storagePolicyName, clientGroupName,
                                schedulePolicyName, locationName, providerDomainName,
                                alertName, workflowName, policyName, roleName

                entity_name:    client name for entity_type 'clientName'
                                Media agent name for entitytype 'mediaAgentName'
                                similar for other entity_typees

                request_type        --      decides whether to ADD, DELETE or
                                            OVERWRITE user security association.

        """
        complete_association = []
        for entity_value in entity_dictionary.values():
            for each_entity_key in entity_value:
                for element in entity_value[each_entity_key]:
                    if each_entity_key != "role":
                        if each_entity_key == "_type_":
                            association_blob = {
                                "entities": {
                                    "entity": [{
                                        each_entity_key: element,
                                        "flags": {
                                            "includeAll": True
                                        }
                                    }]
                                },
                                "properties": {
                                    "role": {
                                        "roleName": entity_value['role'][0]
                                    }
                                }
                            }
                        else:
                            association_blob = {
                                "entities": {
                                    "entity": [{
                                        each_entity_key: element
                                    }]
                                },
                                "properties": {
                                    "role": {
                                        "roleName": entity_value['role'][0]
                                    }
                                }
                            }
                        complete_association.append(association_blob)
        return complete_association

    @staticmethod
    def fetch_security_association(security_dict):
        """Fetches security associations from entity
        Args:
            security_dict    (dict)   --  security association properties of entity

        Returns:
            formatted security association dictionary with custom permissions marked as invalid
        """
        security_list = []
        count = 0
        associations = {}
        entity_permissions = {}
        for every_association in security_dict:
            if 'entity' in every_association['entities']:
                entities = every_association['entities']['entity']
                for entity in entities:
                    for each_key in entity:
                        if 'Name' in each_key:
                            if 'externalGroupName' in each_key:
                                associations = entity[each_key]
                            #if 'providerDomainName' in each_key:
                                # No need to explicitely  check for provider key
                                if associations:
                                    ext_group = "{0}\\{1}".format(entity['providerDomainName'],
                                                                  associations)
                                    associations = {}
                                else:
                                    ext_group = entity[each_key]
                                security_list.append(each_key)
                                security_list.append(ext_group.lower())
                                break
                            else:
                                security_list.append(each_key)
                                security_list.append(entity[each_key].lower())
                                break
                        elif 'flags' in each_key:
                            security_list.append(entity['_type_'])
                            security_list.append(entity['flags'])

                    if 'role' in every_association['properties']:
                        role_list = every_association['properties']['role']
                        for entity in role_list:
                            if 'Name' in entity:
                                security_list.append(role_list[entity].lower())
                    if 'categoryPermission' in every_association['properties']:
                        categories = every_association['properties'][
                            'categoryPermission']['categoriesPermissionList']
                        for key in categories:
                            categories = key
                            for permission in categories:
                                if 'Name' in permission:
                                    security_list.append(categories[permission] + str('-invalid'))
                                    #Not supporting custom permissions as of now.
                    entity_permissions.setdefault(count, security_list)
                    security_list = []
                    count += 1
        return entity_permissions


    def _get_security_roles(self):
        """Returns the list of available roles on this commcell"""
        GET_SECURITY_ROLES = self._commcell_object._services['GET_SECURITY_ROLES']

        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET', GET_SECURITY_ROLES
        )

        if flag:
            if response.json() and 'roleProperties' in response.json():
                role_props = response.json()['roleProperties']

                roles = {}

                for role in role_props:
                    if 'role' in role:
                        role_name = role['role']['roleName'].lower()
                        role_id = role['role']['roleId']
                        roles[role_name] = role_id

                return roles
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def _add_security_association(self, association_list, user= True, request_type = None, externalGroup = False):
        """
        Adds the security association on the specified class object

        Supported Types : Client, Storage Pool class objects.

        Args:
            associations_list   (list)  --  list of users to be associated
                Example:
                    associations_list = [
                        {
                            'user_name': user1,
                            'role_name': role1
                        },
                        {
                            'user_name': user2,
                            'role_name': role2
                        }
                    ]
 
            user (bool)             --    True or False. set user = False, If associations_list made up of user groups
            request_type (str)      --    eg : 'OVERWRITE' or 'UPDATE' or 'DELETE', Default will be OVERWRITE operation
            externalGroup (bool)    --    True or False, set externalGroup = True. If Security associations is to be done on External User Groups

        Raises:
            SDKException:
                if association is not of dict type
                if role doesnot exists on Commcell
                if request fails
        """

        update_operator_request_type = {
            "OVERWRITE": 1,
            "UPDATE": 2,
            "DELETE": 3
        }

        if request_type:
            request_type = request_type.upper()

        security_association_list = []
        for association in association_list:
            if not isinstance(association, dict):
                raise SDKException('Security', '101')

            if not self.has_role(association['role_name']):
                raise SDKException(
                    'Security', '102', 'Role {0} doesn\'t exist'.format(association['role_name'])
                )

            user_or_group = {}
            if user:
                user_or_group = {'userName': association['user_name']}
            elif externalGroup:
                user_or_group = {'externalGroupName': association['user_name']}
            else:
                user_or_group = {'userGroupName': association['user_name']}  

            temp = {
                "userOrGroup": [
                    user_or_group
                ],
                "properties": {
                    "role": {
                        "_type_": 120,
                        "roleId": self._roles[association['role_name'].lower()],
                        'roleName': association['role_name']
                    }
                }
            }
            security_association_list.append(temp)

        request_json = {
            "entityAssociated": self._entity_list,
            "securityAssociations": {
                "associationsOperationType": update_operator_request_type.get(request_type, 1),
                "associations": security_association_list,
                "ownerAssociations": {
                    "ownersOperationType": 1
                }
            }
        }

        ADD_SECURITY_ASSOCIATION = self._commcell_object._services['SECURITY_ASSOCIATION']

        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'POST', ADD_SECURITY_ASSOCIATION, request_json
        )

        if flag:
            if response.json() and 'response' in response.json():
                response_json = response.json()['response'][0]

                error_code = response_json['errorCode']

                if error_code != 0:
                    error_message = response_json['errorString']
                    raise SDKException(
                        'Security',
                        '102',
                        'Failed to add associations. \nError: {0}'.format(error_message)
                    )
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def has_role(self, role_name):
        """Checks if role with specified name exists

            Args:
                role_name     (str)     --     name of the role to be verified

            Returns:
                (bool)     -  True if role with specified name exists
        """
        if not isinstance(role_name, str):
            raise SDKException('Security', '101')

        return self._roles and role_name.lower() in self._roles




# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a Google Cloud Spanner Instance.

GoogleSpannerInstance is the only class defined in this file.

GoogleSpannerInstance: Derived class from CloudAppsInstance Base class, representing a
Google Cloud Spanner instance, and to perform operations on that instance

GoogleSpannerInstance:

    _get_instance_properties()  --  Instance class method overwritten to add cloud apps
    instance properties as well

GoogleSpannerInstance Attributes:

    instance_type           --  Returns the GoogleSpannerInstance instance type

    spanner_instance_id     --  Returns the cloud service account client id

    proxy_client            --  Returns the proxy client name to this instance

    staging_path            --  Returns the instance staging path

    project_id              --  Returns the cloud spanner project id

"""
from ...exception import SDKException
from ..cainstance import CloudAppsInstance


class GoogleSpannerInstance(CloudAppsInstance):
    """Class for representing an Instance of the Google Cloud Spanner instance type."""

    def _get_instance_properties(self):
        """Gets the properties of this instance.

            Raises:
                SDKException:
                    if response is empty

                    if response is not success

        """
        super(GoogleSpannerInstance, self)._get_instance_properties()
        self._ca_instance_type = None
        self._proxy_client = None

        self._google_instance_id = None
        self._staging_path = None
        self._project_id = None

        if 'cloudAppsInstance' in self._properties:
            cloud_apps_instance = self._properties['cloudAppsInstance']
            self._ca_instance_type = cloud_apps_instance['instanceType']

            if 'generalCloudProperties' in cloud_apps_instance:
                if 'proxyServers' in cloud_apps_instance['generalCloudProperties']:
                    self._proxy_client = cloud_apps_instance.get(
                        'generalCloudProperties', {}).get('proxyServers', [{}])[0].get('clientName')
                else:
                    if 'clientName' in cloud_apps_instance.get(
                        'generalCloudProperties', {}).get('memberServers', [{}])[0].get('client'):
                        self._proxy_client = cloud_apps_instance.get('generalCloudProperties', {}).get(
                            'memberServers', [{}])[0].get('client', {}).get('clientName')
                    else:
                        self._proxy_client = cloud_apps_instance.get('generalCloudProperties', {}).get(
                            'memberServers', [{}])[0].get('client', {}).get('clientGroupName')

                if self._proxy_client is None:
                    raise SDKException('Instance', '102', 'Access Node has not been configured')

            if 'cloudSpannerInstance' in cloud_apps_instance:
                self._google_instance_id = cloud_apps_instance.get(
                    'cloudSpannerInstance', {}).get('instanceId')
                self._staging_path = cloud_apps_instance.get(
                    'cloudSpannerInstance', {}).get('cloudStagingPath')
                self._project_id = cloud_apps_instance.get(
                    'cloudSpannerInstance', {}).get('projectId')

    @property
    def instance_type(self):
        """
            Returns:

                 int:   CloudAppsInstance instance type

        """
        return self._ca_instance_type

    @property
    def spanner_instance_id(self):
        """
            Returns:

                 str:   Google service account instance id

        """
        return self._google_instance_id

    @property
    def proxy_client(self):
        """
            Returns the proxy client name to this instance

                str:    Client name of proxy associated to cloud account

        """
        return self._proxy_client

    @property
    def staging_path(self):
        """
            Returns the instance staging path

                str:    Cloud spanner staging path of instance

        """
        return self._staging_path

    @property
    def project_id(self):
        """
            Returns the cloud project id

                str:    Cloud spanner project id
        """
        return self._project_id



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a Cloud Apps Subclient.

CloudAppsSubclient is the only class defined in this file.

CloudAppsSubclient: Derived class from Subclient Base class, representing a
cloud apps subclient, and to perform operations on that subclient.

Note: GoogleSubclient class is used for OneDrive subclient too.

CloudAppsSubclient:

    __new__()   --  Method to create object based on specific cloud apps instance type

"""

from __future__ import unicode_literals

from ..subclient import Subclient
from ..exception import SDKException


class CloudAppsSubclient(Subclient):
    """Class for representing a subclient of the Cloud Apps agent."""

    def __new__(cls, backupset_object, subclient_name, subclient_id=None):
        from .cloudapps.salesforce_subclient import SalesforceSubclient
        from .cloudapps.google_subclient import GoogleSubclient
        from .cloudapps.cloud_storage_subclient import CloudStorageSubclient
        from .cloudapps.cloud_database_subclient import CloudDatabaseSubclient
        from .cloudapps.dynamics365_subclient import MSDynamics365Subclient
        from .cloudapps.teams_subclient import TeamsSubclient
        from .cloudapps.spanner_subclient import GoogleSpannerSubclient

        instance_types = {
            1: GoogleSubclient,
            2: GoogleSubclient,
            3: SalesforceSubclient,
            4: CloudDatabaseSubclient,  # Amazon RDS Subclient
            5: CloudStorageSubclient,  # AmazonS3 Subclient
            6: CloudStorageSubclient,  # AzureBlob Subclient
            7: GoogleSubclient,  # OneDrive Subclient. GoogleSuclient class is used for OneDrive too
            14: CloudStorageSubclient,  # OracleCloud Subclient
            15: CloudStorageSubclient,  # Openstack Subclient
            20: CloudStorageSubclient,  # Google Cloud Instance
            21: CloudStorageSubclient,  # azure data lake gen2
            26: CloudDatabaseSubclient,  # Amazon Redshift subclient
            27: CloudDatabaseSubclient,  # Amazon Document DB subclient
            25: CloudStorageSubclient,  # AliBaba
            24: CloudStorageSubclient,  # IBM
            22: CloudDatabaseSubclient,  # Amazon DynamoDB subclient
            35: MSDynamics365Subclient,  # Office 365 Apps -> MS Dynamics 365 Subclient
            36: TeamsSubclient,  # Office 365 Apps -> MS Teams
            37: GoogleSpannerSubclient  # Google Cloud Spanner Subclient
        }

        cloud_apps_instance_type = backupset_object._instance_object._properties[
            'cloudAppsInstance']['instanceType']

        if cloud_apps_instance_type in instance_types:
            instance_type = instance_types[cloud_apps_instance_type]
        else:
            raise SDKException('Subclient', '112')

        return object.__new__(instance_type)



# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------
"""
SharepointInstance is the only class defined in this file.

SharepointInstance  :  Derived class from Instance  Base class, representing a
                       Sharepoint Instance, and to perform operations on that instance

SharepointInstance:

     _restore_common_options_json()  -- setter for common options property in restore

     _restore_json() --  Method which creates json for a restore job

     _restore_v1_json() -- Method which creates json for v1 client for a restore job

"""

from cvpysdk.exception import SDKException
from ..instance import Instance



class SharepointInstance(Instance):
    """ Class  representing a sharepoint instance, and to perform operations on that instance
    """

    def _restore_browse_option_json(self, value):
        """setter for the Browse options for restore in Json"""

        if not isinstance(value, dict):
            raise SDKException('Instance', '101')

        time_range_dict = {}
        if value.get('to_time'):
            time_range_dict['toTime'] = value.get('to_time')
        self._browse_restore_json = {
            "commCellId":  int(self._commcell_object.commcell_id),
            "showDeletedItems": value.get("showDeletedItems", False),
            "backupset": {
                "clientName": self._agent_object._client_object.client_name,
                "appName": self._agent_object.agent_name,
                "clientId": int(self._instance['clientId']),
                "backupsetId": int(self._restore_association['backupsetId'])
            },
            "timeRange": time_range_dict
        }

    def _restore_common_options_json(self, value):
        """setter for  the Common options of in restore JSON"""
        if not isinstance(value, dict):
            raise SDKException('Instance', '101')

        self._commonoption_restore_json = {
            "allVersion": True,
            "offlineMiningRestore": False,
            "skip": not value.get("unconditional_overwrite", False),
            "restoreACLs": False,
            "erExSpdbPathRestore": True,
            "unconditionalOverwrite": value.get("unconditional_overwrite", False),
            "siteReplicationrestore": False,
            "append": False
        }

    def _restore_destination_json(self, value):
        """setter for  the destination restore option in restore JSON"""
        if not isinstance(value, dict):
            raise SDKException('Subclient', '101')

        self._destination_restore_json = {
            "inPlace": value.get("in_place", True),
            "destClient": {
                "clientName": value.get("client_name", ""),
                "clientId": value.get("client_id", -1)
            }
        }

    def _restore_json(self, **kwargs):
        """
        Creates json required for restore job

        Kwargs:

            paths   (list)  --  list of sites or webs to be restored
                Example : [
                        "MB\\https://cvdevtenant.sharepoint.com/sites/TestSite\\/\\Shared Documents\\TestFolder",
                        "MB\\https://cvdevtenant.sharepoint.com/sites/TestSite\\/\\Lists\\TestList",
                        ]
        Returns:

            rest_json  (dict)  --  dictionary with parameters set required for a restore job

        """
        if(kwargs.get("v1",False)):
            return self._restore_v1_json(**kwargs)
        rest_json = super(SharepointInstance, self)._restore_json(**kwargs)
        rest_json["taskInfo"]["task"]["initiatedFrom"] = 1
        rest_json["taskInfo"]["subTasks"][0]["options"]["restoreOptions"]["sharePointDocRstOption"] = {}
        rest_json["taskInfo"]["subTasks"][0]["options"]["restoreOptions"]\
            ["sharePointRstOption"]= {
                "sharePointDocument": True,
                "spRestoreToDisk": {
                    "restoreToDiskPath": "",
                    "restoreToDisk": False
                }
            }
        rest_json["taskInfo"]["subTasks"][0]["options"]["commonOpts"] = {
            "notifyUserOnJobCompletion": False
        }
        return rest_json

    def _restore_v1_json(self, **kwargs):
        """
        Creates json required for restore job for v1 client

        Kwargs:

            paths   (list)  --  list of sites or webs to be restored
                Example : [
                        "MB\\https://cvdevtenant.sharepoint.com/sites/TestSite\\/\\Shared Documents\\TestFolder",
                        "MB\\https://cvdevtenant.sharepoint.com/sites/TestSite\\/\\Lists\\TestList",
                        ]
        Returns:

            rest_json  (dict)  --  dictionary with parameters set required for a restore job

        """
        restore_option = {}
        if kwargs.get("restore_option"):
            restore_option = kwargs["restore_option"]
            for key in kwargs:
                if not key == "restore_option":
                    restore_option[key] = kwargs[key]
        else:
            restore_option.update(kwargs)

        if self._restore_association is None:
            self._restore_association = self._instance

        if restore_option.get('copy_precedence') is None:
            restore_option['copy_precedence'] = 0

        if restore_option.get('overwrite') is not None:
            restore_option['unconditional_overwrite'] = restore_option['overwrite']

        if restore_option.get('live_browse'):
            restore_option['liveBrowse'] = True
        else:
            restore_option['liveBrowse'] = False

        # restore_option should use client key for destination client info
        client = restore_option.get("client", self._agent_object._client_object)

        if isinstance(client, str):
            client = self._commcell_object.clients.get(client)

        restore_option["client_name"] = client.client_name
        restore_option["client_id"] = int(client.client_id)

        # set time zone
        from_time = restore_option.get("from_time", None)
        to_time = restore_option.get("to_time", None)
        time_list = ['01/01/1970 00:00:00', '1/1/1970 00:00:00']

        if from_time and from_time not in time_list:
            restore_option["from_time"] = from_time

        if to_time and to_time not in time_list:
            restore_option["to_time"] = to_time

        self._restore_browse_option_json(restore_option)
        self._restore_common_options_json(restore_option)
        self._restore_destination_json(restore_option)
        self._restore_fileoption_json(restore_option)
        self._restore_common_opts_json(restore_option)

        if not restore_option.get('index_free_restore', False):
            if restore_option.get("paths") == []:
                raise SDKException('Subclient', '104')

        request_json = {
            "taskInfo": {
                "associations": [self._restore_association],
                "task": {
                    "taskType": 1,
                    "initiatedFrom": 1
                },
                "subTasks": [{
                    "subTask": {
                        "subTaskType": 3,
                        "operationType": 1001
                    },
                    "options": {
                        "restoreOptions": {
                            "sharePointDocRstOption": {
                                "isWorkflowAlertsRestoreOnly": False
                            },
                            "browseOption": self._browse_restore_json,
                            "commonOptions": self._commonoption_restore_json,
                            "destination": self._destination_restore_json,
                            "fileOption": self._fileoption_restore_json,
                            "sharePointRstOption": {
                                "sharePointDocument": True,
                                "spRestoreToDisk": {
                                    "restoreToDiskPath": "",
                                    "restoreToDisk": False
                                }
                            },
                        },
                        "commonOpts": self._commonopts_restore_json
                    }
                }]
            }
        }
        return request_json

